# exoplanet population of all "confirmed" exoplanets from exoplanet archive
from ...imports import *
from ..predefined import PredefinedPopulation
from .exoplanet_downloaders import *

__all__ = ["Exoplanets", "ExoplanetsPSCP", "ExoplanetsPS"]

suffixes = [
    "_reference",
    "",
    "_uncertainty_lower",
    "_uncertainty_upper",
    "_lower_limit",
    "_upper_limit",
]


def parse_reflink(x):
    """
    Convert a 'reflink/refname' from NASA Exoplanet Archive into
    a human-friendly string and a definitely-unique URL.

    Parameters
    ----------
    x : str
        The reflink or refname, which might look like:
        '<a refstr=TORRES_ET_AL__2008 href=https://ui.adsabs.harvard.edu/abs/2008ApJ...677.1324T/abstract target=ref> Torres et al. 2008 </a>'

    Returns
    -------
    name : str
        The human-friendly string for this reference. Be careful, this might not
        necessarily by 100% unique, particularly if folks publish multiple planets
        in the same year.
    """
    try:
        url = x.split("href=")[1].split(" ")[0]
        name = x.split(">")[1].split("<")[0].replace("&amp;", "+").strip()
        return name  # , url
    except AttributeError:
        return ""


class ExoplanetsPSCP(PredefinedPopulation):
    """
    Exoplanets from the NASA Exoplanet Archive.

    This uses the Planetary Systems Composite Parameters table,
    which merges together data from different papers for the
    same planet, which can sometimes result in parameters
    that aren't necessarily consistent with each other.
    """

    label = "ExoplanetsPSCP"

    # define columns to ingest with and without errors/limits
    _raw_columns_with_errors_and_limits = [
        "pl_orbper",
        "pl_orbsmax",
        "pl_rade",
        "pl_bmasse",
        "pl_msinie",
        "pl_dens",
        "pl_orbeccen",
        "pl_orbincl",
        "pl_orblper",
        "pl_tranmid",
        "pl_trandep",
        "pl_trandur",
        "pl_imppar",
        "pl_ratdor",
        "pl_ratror",
        "pl_rvamp",
        "pl_projobliq",
        "pl_trueobliq",
        "pl_insol",
        "pl_eqt",
        "st_teff",
        "st_rad",
        "st_mass",
        "st_met",
        "st_lum",
        "st_logg",
        "st_age",
        "st_dens",
        "st_vsin",
        "st_rotp",
        "st_radv",
    ]
    _raw_columns_with_errors = [
        "sy_dist",
        "sy_plx",
        "sy_bmag",
        "sy_vmag",
        "sy_jmag",
        "sy_hmag",
        "sy_kmag",
        "sy_umag",
        "sy_gmag",
        "sy_rmag",
        "sy_imag",
        "sy_zmag",
        "sy_w1mag",
        "sy_w2mag",
        "sy_w3mag",
        "sy_w4mag",
        "sy_gaiamag",
        "sy_tmag",
        "sy_kepmag",
        "sy_icmag",
    ]
    _raw_columns_without_errors = [
        "pl_name",
        "hostname",
        "pl_letter",
        "ra",
        "dec",
        "sy_pmra",
        "sy_pmdec",
        "gaia_id",
        "sy_snum",
        "sy_pnum",
        "discoverymethod",
        "disc_year",
        "disc_refname",
        "disc_facility",
        "rv_flag",
        "pul_flag",
        "ptv_flag",
        "tran_flag",
        "ast_flag",
        "obm_flag",
        "micro_flag",
        "etv_flag",
        "ima_flag",
        "dkin_flag",
        "pl_controv_flag",
        "ttv_flag",
        "st_spectype",
        "pl_nespec",
        "pl_ntranspec",
        "default_flag",
    ]
    _downloader = ExoplanetArchiveDownloader("pscomppars")

    def __init__(self, remake=False, **plotkw):
        """
        Initialize a population of all known exoplanets
        using the NASA Exoplanet Archive `pscomppars` table.

        This generates an Exoplanets population containing
        planets discovered through any method. It tries to
        be clever about loading cached files so that it
        will work quickly and not try to redownload everything
        from scratch every time it's called.

        Parameters
        ----------
        remake : bool
            Should the population be remade from raw files,
            even if a recent version already exists?
        **plotkw : dict
            All other keywords will go toward defining
            default plotting styles, for example like
            'alpha', 'marker', 'zorder', 'color', ...
        """

        # load standard table(s) or ingest from raw data
        PredefinedPopulation.__init__(self, remake=remake, **plotkw)
        # self._add_references_as_indices()

    def _add_references_as_indices(self):
        """
        Add all keys that look like references as table indices for faster lookup.
        """
        for k in self.standard.colnames:
            if "_reference" in k:
                self._make_sure_index_exists(k)

    def _create_standardized(self, raw):
        """
        Create a standardized table to make sure that at
        least a few necessary columns are populated.

        This standardization step aims to make it easier to
        compare exoplanet populations from different sources.
        It translates columns from a raw table into a standard
        set of expected column names, doing some calculations
        if necessary.

        Parameters
        ----------
        raw : astropy.table.QTable
            A raw unstandardized table from `ExoplanetArchiveDownloader`.

        Returns
        -------
        standard : astropy.table.QTable
            A standardize table of exoplanet properties.
        """

        # define the table from which we're deriving everying
        r = raw

        # the new standardized table
        s = QTable()

        def strip_even_if_masked(x):
            """
            Flexibly clean strings by removing trailing spaces.
            """
            try:
                return x.strip()
            except AttributeError:
                return None

        def attach_unit(x, unit=None):
            """
            Flexibly give units to quantities that should have them.
            """
            if isinstance(x[0], str):
                return [strip_even_if_masked(a) for a in x]
            if unit is None:
                return x
            else:
                return x * unit

        def populate_one_or_more_columns(k_new, k_original, unit=None):
            """
            A wrapper to help smoothly grab quantities from raw
            table and ingest them into the standardized table.

            For columns with

            Parameters
            ----------
            k_original : string
                The column name in the original, raw table.
            k_new : string
                The column name in the new, standardized table.
            unit : None, astropy.units.Unit
                A unit to attach to the quantity, if necessary.
            """

            if k_original not in r.colnames:
                warnings.warn(f"⚠️ No {k_original} found!")
                return

            if k_original in self._raw_columns_without_errors:
                # easy, just record the column itself with no error
                s[k_new] = attach_unit(r[k_original], unit)
                print(f"📕 populated {k_original} > {k_new}")

            elif k_original in self._raw_columns_with_errors:
                # record the column itself
                s[k_new] = attach_unit(r[k_original], unit)

                # record the upper and lower errorbars
                s[f"{k_new}_uncertainty_upper"] = attach_unit(
                    r[f"{k_original}err1"], unit
                )
                s[f"{k_new}_uncertainty_lower"] = attach_unit(
                    r[f"{k_original}err2"], unit
                )
                print(f"📏 populated {k_new} and errors with {k_original}")

            elif k_original in self._raw_columns_with_errors_and_limits:
                # from playing with table, I think lim = +1 is upper limit, -1 is lower limit

                # record the upper and lower errorbars
                limit_flag = r[f"{k_original}lim"]

                # populate lower limits
                is_lower = limit_flag == -1
                if np.sum(is_lower) > 0:
                    s[f"{k_new}_lower_limit"] = attach_unit(r[f"{k_original}"])
                    s[f"{k_new}_lower_limit"][is_lower == False] = np.nan

                # populate upper limits
                is_upper = limit_flag == +1
                if np.sum(is_upper) > 0:
                    s[f"{k_new}_upper_limit"] = attach_unit(r[f"{k_original}"])
                    s[f"{k_new}_upper_limit"][is_upper == False] = np.nan

                # populate bounded constraints
                is_bounded = limit_flag == 0
                if np.sum(is_bounded) > 0:
                    s[k_new] = attach_unit(r[k_original], unit)
                    s[f"{k_new}_uncertainty_upper"] = attach_unit(
                        r[f"{k_original}err1"], unit
                    )
                    s[f"{k_new}_uncertainty_lower"] = attach_unit(
                        r[f"{k_original}err2"], unit
                    )
                    s[k_new][is_bounded == False] = np.nan
                    s[f"{k_new}_uncertainty_upper"][is_bounded == False] = np.nan
                    s[f"{k_new}_uncertainty_lower"][is_bounded == False] = np.nan

                    print(f"👇 populated {k_original} > {k_new} and errors and limits ")

            else:
                # easy, just record the column itself with no error
                s[k_new] = attach_unit(r[k_original], unit)
                print(f"🙋 populated {k_original} > {k_new} , but not 100% sure...")

            # keep track of reference for measurements
            if (k_original in self._raw_columns_with_errors_and_limits) or (
                k_original in self._raw_columns_with_errors
            ):
                try:
                    s[f"{k_new}_reference"] = self._ingest_references(r, k_original)
                    print(
                        f"⚠️ ingested reference information for {k_original} > {k_new}"
                    )
                    # s.add_index(f"{k_new}_reference")
                except (KeyError, AssertionError):
                    print(
                        f"⚠️ no reference information found for {k_original} > {k_new}"
                    )

        # basic reference information
        populate_one_or_more_columns("name", "pl_name")
        populate_one_or_more_columns("hostname", "hostname")
        populate_one_or_more_columns("letter", "pl_letter")
        populate_one_or_more_columns("gaia_id", "gaia_id")
        populate_one_or_more_columns("number_of_stars", "sy_snum")
        populate_one_or_more_columns("number_of_planets", "sy_pnum")

        # what's the history?
        populate_one_or_more_columns("discovery_method", "discoverymethod")
        populate_one_or_more_columns("discovery_year", "disc_year")
        populate_one_or_more_columns("discovery_publication", "disc_refname")
        s["discovery_publication"] = [
            parse_reflink(x) for x in s["discovery_publication"]
        ]
        populate_one_or_more_columns("discovery_facility", "disc_facility")

        # what are the host positions and kinematics?
        populate_one_or_more_columns("ra", "ra", u.deg)
        populate_one_or_more_columns("dec", "dec", u.deg)
        populate_one_or_more_columns("pmra", "sy_pmra", u.mas / u.year)
        populate_one_or_more_columns("pmdec", "sy_pmdec", u.mas / u.year)
        populate_one_or_more_columns("systemic_rv", "st_radv", u.km / u.s)
        populate_one_or_more_columns("distance", "sy_dist", u.pc)

        # what are some useful flags?
        populate_one_or_more_columns("detected_in_rv", "rv_flag")
        populate_one_or_more_columns("detected_in_pulsar", "pul_flag")
        populate_one_or_more_columns("detected_in_pulsation_timing", "ptv_flag")
        populate_one_or_more_columns("detected_in_transit", "tran_flag")
        populate_one_or_more_columns("detected_in_astrometry", "ast_flag")
        populate_one_or_more_columns(
            "detected_in_orbital_brightness_modulations", "obm_flag"
        )
        populate_one_or_more_columns("detected_in_microlensing", "micro_flag")
        populate_one_or_more_columns(
            "detected_in_eclipse_timing_variations", "etv_flag"
        )
        populate_one_or_more_columns("detected_in_imaging", "ima_flag")
        populate_one_or_more_columns("detected_in_disk_kinematics", "dkin_flag")

        populate_one_or_more_columns("is_controversial", "pl_controv_flag")
        populate_one_or_more_columns("shows_ttv", "ttv_flag")

        # (these might need merging from the composite table?!)
        populate_one_or_more_columns("stellar_spectral_type", "st_spectype")
        populate_one_or_more_columns("stellar_teff", "st_teff", u.K)
        populate_one_or_more_columns("stellar_radius", "st_rad", u.Rsun)
        populate_one_or_more_columns("stellar_mass", "st_mass", u.Msun)
        populate_one_or_more_columns("stellar_age", "st_age", u.Gyr)
        populate_one_or_more_columns("stellar_metallicity", "st_met")
        populate_one_or_more_columns("stellar_luminosity", "st_lum")
        # convert from table log10(L) to L
        logL = s["stellar_luminosity"] * 1
        L = 10**logL * u.Lsun
        logL_uncertainty_lower = s["stellar_luminosity_uncertainty_lower"] * 1
        logL_uncertainty_upper = s["stellar_luminosity_uncertainty_upper"] * 1
        s["stellar_luminosity"] = L
        s["stellar_luminosity_uncertainty_lower"] = (
            np.log(10) * L * np.abs(logL_uncertainty_lower)
        )
        s["stellar_luminosity_uncertainty_upper"] = (
            np.log(10) * L * np.abs(logL_uncertainty_upper)
        )

        populate_one_or_more_columns("stellar_logg", "st_logg")
        populate_one_or_more_columns("stellar_density", "st_dens")
        populate_one_or_more_columns("stellar_vsini", "st_vsin")
        populate_one_or_more_columns("stellar_rotation_period", "st_rotp")

        # what are the stellar magnitudes?
        bands = [
            "u",
            "g",
            "r",
            "i",
            "z",
            "V",
            "B",
            "IC",
            "J",
            "H",
            "K",
            "W1",
            "W2",
            "W3",
            "W4",
            "gaia",
            "T",
            "kep",
        ]
        for b in bands:
            populate_one_or_more_columns(f"magnitude_{b}", f"sy_{b.lower()}mag", u.mag)

        # what are some basic orbital parameters
        populate_one_or_more_columns("period", "pl_orbper", u.day)
        populate_one_or_more_columns("semimajoraxis", "pl_orbsmax", u.AU)
        populate_one_or_more_columns("eccentricity", "pl_orbeccen")
        populate_one_or_more_columns("argument_of_periastron", "pl_orblper", u.deg)
        populate_one_or_more_columns("inclination", "pl_orbincl", u.deg)

        # what are the planet properties? (check Jupiter isn't better?!)
        populate_one_or_more_columns("radius", "pl_rade", u.Rearth)
        populate_one_or_more_columns("mass", "pl_bmasse", u.Mearth)
        populate_one_or_more_columns("density", "pl_dens", u.g / u.cm**3)
        flux_unit = (1 * u.Lsun / 4 / np.pi / (1 * u.AU) ** 2).to("W/m**2")
        # populate_one_or_more_columns("insolation", "pl_insol", flux_unit)
        # populate_one_or_more_columns("teq", "pl_eqt", u.K)

        # does it have transmission + emission spec?
        populate_one_or_more_columns(
            "number_of_transmission_measurements", "pl_ntranspec"
        )
        populate_one_or_more_columns("number_of_emission_measurements", "pl_nespec")

        # what are the (often) transit-derived properties?
        populate_one_or_more_columns("transit_midpoint", "pl_tranmid", u.day)
        populate_one_or_more_columns("transit_duration", "pl_trandur", u.hour)
        populate_one_or_more_columns("transit_scaled_radius", "pl_ratror")
        populate_one_or_more_columns("transit_depth", "pl_trandep", 0.01)
        populate_one_or_more_columns("transit_scaled_semimajoraxis", "pl_ratdor")
        populate_one_or_more_columns("transit_impact_parameter", "pl_imppar")

        # what are the (often) RV-derived properties
        populate_one_or_more_columns("rv_semiamplitude", "pl_rvamp", u.m / u.s)
        populate_one_or_more_columns("msini", "pl_msinie", u.Mearth)
        populate_one_or_more_columns("projected_obliquity", "pl_projobliq", u.deg)
        populate_one_or_more_columns("obliquity", "pl_trueobliq", u.deg)

        # is this the default parameter set?
        populate_one_or_more_columns("default_parameter_set", "default_flag")

        # sort these planets by their names
        s.sort("name")

        # set the fill_value for any numeric columns to nans
        for k in s.colnames:
            if not isinstance(s[k][0], str):
                s[k].fill_value = np.nan

        # fill in all the masked elements to make an unmasked array with nans
        standard = s.filled()

        #
        trimmed = self._trim_bad_data(standard)

        # return that standardized table
        return trimmed

    def _trim_bad_data(self, standard):
        """
        Mask bad data from the '.standard' table.

        Planetary Systems Composite Parameters contains lots of quantities
        that aren't quite reliable empircal measurements that we want to use.
        This trims theoretical calculations and/or suspicious measurements
        out of the population, masking them and replacing their central values
        with np.nan.

        Parameters
        ----------
        standard : astropy.table.QTable
            The filled table that will populate the `.standard` table.

        Returns
        -------
        trimmed : astropy.table.QTable
            The .standard table with bad data tidied away.
        """

        print("✂️ trimming weird data")

        # make a copy of the table
        trimmed = copy.deepcopy(standard)

        # find the columns that should have uncertainties
        columns_with_uncertainties = np.unique(
            [
                x.split("_uncertainty")[0]
                for x in standard.colnames
                if "uncertainty" in x
            ]
        )
        for k in columns_with_uncertainties:

            # remove quantities with no or nonsense uncertainties
            ok = np.isfinite(standard[k])
            N_has_value = np.sum(ok)
            for w in ["lower", "upper"]:
                ok *= np.isfinite(standard[f"{k}_uncertainty_{w}"])
                # this comment is a potentially harmless or potentially troublesome KLUDGE!
                # see https://github.com/zkbt/exoatlas/issues/58
                # ok *= standard[f"{k}_uncertainty_{w}"] != 0
                print(
                    f"🚨😳🔔‼️ some values with zero-uncertainty might have snuck through for {k} 🚨😳🔔‼️🚨😳🔔‼️"
                )
            N_has_uncertainty = np.sum(ok)
            bad = ok == False
            print(
                f"""{k:>20}: {N_has_value:>4} values and {N_has_uncertainty:>4} with uncertainties ({N_has_uncertainty/N_has_value:>6.1%})"""
            )
            trimmed[k][bad] = np.nan

        print("values without reasonable uncertainties have been trimmed")
        return trimmed

    def _ingest_references(self, r, k):
        """
        Try to get the reference for a particular measurement.

        For the Planetary Systems Composite Parameters table,
        where references have merged, each measured quantity has
        its own reference encoded as a '*reflink' column name.

        Parameters
        ----------
        r : astropy.table.QTable
            The raw table being ingested.
        k : str
            The key for which a reference should be checked.
        """

        # define a column name for this quantity
        k_reference = f"{k}_reflink"

        # get a bibcode for this quantity from each row
        list_of_references = [parse_reflink(x) for x in r[k_reference]]

        return list_of_references


class ExoplanetsPS(ExoplanetsPSCP):
    """
    Exoplanets from the NASA Exoplanet Archive.

    This uses the Planetary Systems table,
    which includes one row per planet per reference,
    so it's a massive dataset providing multiple options
    for most planets, although any one particular reference
    might be missing the complete set of parameters needed
    to fully characterize any one system.
    """

    label = "ExoplanetsPS"
    _downloader = ExoplanetArchiveDownloader("ps")

    def __init__(self, *args, **kwargs):
        """
        Initialize a population of all known exoplanets
        using the NASA Exoplanet Archive `ps` table.

        This contains many more rows than there are planets,
        because there's one row for each reference (for each planet).

        This also sets all the `*_reference` columns to serve as
        table indices, to make them easy to extract for updating
        individual planet references.
        """
        ExoplanetsPSCP.__init__(self, *args, **kwargs)
        self._add_references_as_indices()

    def _ingest_references(self, r, k):
        """
        Try to get the reference for a particular measurement.

        For the Planetary Systems table, where references are
        necessarily consistent across parameter sets, there
        are only three reference columns:
            'pl_refname' for the planet and orbit
            'st_refname' for the stellar properties
            'sy_refname' for the system positions + motions

        Parameters
        ----------
        r : astropy.table.QTable
            The raw table being ingested.
        k : str
            The key for which a reference should be checked.
        """

        # is this key 'pl', 'st', or 'sy'?
        prefix = k.split("_")[0]
        assert prefix in ["pl", "st", "sy"]

        # define a column name for this quantity
        k_reference = f"{prefix}_refname"

        # get a bibcode for this quantity from each row
        list_of_references = [parse_reflink(x) for x in r[k_reference]]

        return list_of_references


class Exoplanets(ExoplanetsPSCP):
    """
    Exoplanets from the NASA Exoplanet Archive.

    This uses the table of Planetary Systems Composite Parameters
    (one row per planet, references merged) but it secretly also
    loads the table of Planetary Systems (one row per reference,
    with multiple rows per planet), so that we can choose to use
    different references for any particular planet.
    """

    label = "Exoplanets"

    def __init__(self, remake=False, **plotkw):
        """
        Initialize a population of all known exoplanet from
        the NASA Exoplanet Archive `pscomppars` and `ps` tables.

        The standard population comes only from `pscomppars`.
        However, it should secretly also download the `ps` table,
        which would enable the user to choose to replace
        parameters for a particular planet by choosing a
        particular reference.

        Parameters
        ----------
        remake : bool
            Should the population be remade from raw files,
            even if a recent version already exists?
        **plotkw : dict
            All other keywords will go toward defining
            default plotting styles, for example like
            'alpha', 'marker', 'zorder', 'color', ...
        """

        # load standard table(s) or ingest from raw data
        PredefinedPopulation.__init__(self, remake=remake, **plotkw)

        # plotting defaults
        self.s = 6
        self.marker = "."
        self.zorder = 0
        self.color = "black"
        self.respond_to_color = True
        self.exact = False

    def __getitem__(self, key):
        """
        Create a subpopulation of planets by indexing, slicing, or masking.

        `Exoplanets` needs its own version of this (beyond the one inherited from
        the overarching `Population` class) because it secretly contains a hidden
        `ExoplanetsPS` table that should also be trimmed down to only relevant
        rows every time that we select a subset.

        Parameters
        ----------
        key : int, list, array, slice
            An index, slice, or mask to select a subset of the population.

        Returns
        -------
        subset : Population
            A subset of the population, as set by the index, slice, or mask.
        """

        # first call the normal `Population` indexing/slicing/masking
        subset = super().__getitem__(key)

        # then trim the .ps population to only those (host)names in the subset
        if hasattr(self, "individual_references"):
            subset.individual_references = self.individual_references[
                list(np.unique(subset.tidyname()))
            ]
            subset.individual_references.label = "Individual References"

        return subset

    def load_individual_references(self):
        """
        Populate an internal `.individual_references` population
        with individual references for planets.

        The `pscomppars` table from which the Exoplanets population
        is defined merged together multiple different references.
        Sometimes we might want be able to see and/or choose the
        values associated with one particular reference. This
        function loads a separate Exoplanet-like population
        containing those individual references and stores it
        in the `.individual_references` attribute.
        """

        # skip slowly loading the data if it already exists
        try:
            return self.individual_references
        except AttributeError:

            print("(loading individual references may take up to a few minutes)")

            # load the internal population with data for individual references
            self.individual_references = ExoplanetsPS()
            # then trim the .ps population to only those (host)names in the subset
            self.individual_references = self.individual_references[
                list(np.unique(self.tidyname()))
            ]

            self.individual_references.label = "Individual References"
            return self.individual_references

    def plot_measurements_from_individual_references(self, x="mass", y="radius"):
        """
        Plot all available measurements from individual references.

        Parameters
        ----------
        x : str
            The quantity to plot on the x-axis.
        y : str
            The quantity to plot on the y-axis.
        """
        for p, c in zip([self.individual_references, self], ["orchid", "black"]):
            plt.errorbar(
                p.get(x),
                p.get(y),
                p.get_uncertainty_lowerupper(y),
                p.get_uncertainty_lowerupper(x),
                linewidth=0,
                elinewidth=1,
                marker="o",
                color=c,
                label=p.label,
            )
        plt.xlabel(x)
        plt.ylabel(y)
        plt.legend()

    def display_individual_references(
        self,
        keys=["mass", "radius", "stellar_mass", "stellar_radius", "stellar_teff"],
        planets=None,
    ):
        """
        Summarize the available measurements for a particular quantity.

        Parameters
        ----------
        planets : str, list
            One planet to investigate, as a string,
            or multiple planets to investigate, as a list of strings,
            or, if None, show the entire population.
        keys : str, list
            One quantity to investigate, as a string,
            or multiple quantities as a list of strings.

        Returns
        -------
        table : Table
            An abbreviated table summarizing the references and the
            measurements they each provide, for the requested key(s).
            The table will be sorted by planet name, to facilitate
            comparing values within a particular system.
        """

        # extract the rows relevant to the requested planets
        if planets == None:
            pop = self
        else:
            pop = self[planets]

        # replace a single key with a list of keys
        if isinstance(keys, str):
            keys = [keys]

        # construct a list of columsn to appear in the final table
        columns_to_include = ["name"]

        # add all columns relevant to the requested key(s)
        available_colnames = pop.standard.colnames
        for k in keys:
            if f"{k}_reference" in available_colnames:
                for suffix in suffixes:
                    if f"{k}{suffix}" in available_colnames:
                        columns_to_include += [f"{k}{suffix}"]

        # construct a subset table with just those columns, first row as actual, others as all available options
        table = vstack(
            [
                pop.standard[columns_to_include],
                pop.individual_references.standard[columns_to_include],
            ]
        )
        table.add_column(["🏷️"] * len(table), index=0, name="is_being_used")
        table["is_being_used"][: len(pop)] = "🗺️"
        table.sort("name")

        return table

    def update_reference(self, references, keys=None, planets=None, verbose=False):
        """
        Modify the internal standardized data table to use
        measurements from a particular reference.

        Each planet may have multiple measurements available in the
        NASA Exoplanet Archive. If you want to switch from the default
        parameters that come from the merged Planetary Systems
        Composite Parameters table to those coming from a particular
        individual reference in the Planetary Systems table, you might...

        1.  Run `.load_individual_references()` to make sure that the
            data for individual references are loaded and stored in the
            `.individul_references` attribute.
        2.  Run `.display_individual_references()` to see what references
            have useful data you might want to use for a particular planet.
        3.  Run `.update_reference()` to use a reference to update the
            parameters in the standardized population table.

        Running this method will permanently modify the `.standard` table
        inside the `Exoplanets` population. The only way to be sure to
        undo the effect is to create a new `Exoplanets` population.

        This will only update finite values; it won't overwrite existing
        values with nan. To do so, we either need to add a keyword option
        here, or you'll need to use `.update_values`.

        TO-DO: This feels slow and inefficient; I'm sure there's a faster
        and/or classier way to do this, but let's call this good for now!

        Parameters
        ----------
        references : str, list
            One reference to try to use, as a string,
            or multiple references to try to use, as a list of strings.
            References will be updated in order, so later references
            in a list will overwrite earlier ones.
        keys : str, list
            One quantity to try to update, as a string,
            or multiple quantities as a list of strings.
        planets : str, list
            One planet to try to update, as a string,
            or multiple planets to update, as a list of strings,
            or, if None, consider the entire population.
        verbose : bool
            Whether all changes should be displayed.
        """

        warning_message = f"""
        Your request to update references with...
            references = {references}
            keys = {keys}
            planets = {planets}
        ...failed somehow. Sorry!
        """
        # make sure references are a list, even with just one element
        if isinstance(references, str):
            references = [references]

        # make sure keys are a list, which might be all columns with references
        if keys == None:
            keys_with_reference = [
                x.split("_reference")[0]
                for x in self.standard.colnames
                if "_reference" in x
            ]
            keys = [k for k in keys_with_reference if k in self.standard.colnames]
        elif isinstance(keys, str):
            keys = [keys]

        # make sure planets are a list, which might be all planets in population
        if planets == None:
            planets_to_index = slice(None, None, None)
        elif isinstance(planets, str):
            planets_to_index = [clean(planets).lower()]
        else:
            planets_to_index = [clean(x).lower() for x in planets]

        # make sure something happens
        nothing_happened = True

        # loop over columns
        for k in keys:

            # extract a table of just these references, or give up
            try:
                # print(f"{k}_reference", references)
                these_references = QTable(
                    self.individual_references.standard.loc[
                        f"{k}_reference", references
                    ]
                )
            except KeyError:
                # if verbose:
                #    print(f'No reference(s) "{references}" found for key "{k}"')
                continue

            # extract a (possibly even smaller) table of just these planets, or give up
            try:
                these_planets = QTable(
                    these_references.loc["tidyname", planets_to_index]
                )
            except KeyError:
                if verbose:
                    print(
                        'No planet(s) "{planets_to_index}" found for reference "{references}"; moving on.'
                    )
                continue

            # find where there is a real new measurement for this key
            is_finite = np.isfinite(these_planets[f"{k}"])
            # if verbose:
            #    print(
            #        k,
            #        planets_to_index,
            #        is_finite,
            #        type(is_finite).__name__,
            #        (type(these_planets).__name__),
            #    )
            new = copy.deepcopy(these_planets[is_finite])
            new.add_index("tidyname")

            # if verbose:
            #    print("The new properties to be adopted:")
            #    display(new)

            # construct a new table of the references we want to update for this key
            if len(new) > 0:
                for tidyname in np.unique(new["tidyname"]):
                    if verbose:
                        print(f"'{k}' for '{tidyname}'")
                    old_for_this_planet = self.standard.loc["tidyname", tidyname]
                    new_for_this_planet = new.loc["tidyname", tidyname]
                    # if verbose:
                    # print("old", old_for_this_planet["tidyname"])
                    # print("new", new_for_this_planet["tidyname"])
                    assert (
                        old_for_this_planet["tidyname"]
                        == new_for_this_planet["tidyname"]
                    )

                    keys_to_display = ["tidyname"]
                    if verbose:
                        old_for_display = copy.deepcopy(old_for_this_planet)

                    for s in suffixes:
                        try:
                            if verbose:
                                old_value = old_for_this_planet[f"{k}{s}"]
                                new_value = new_for_this_planet[f"{k}{s}"]
                                # print(f"{k}{s}: {old_value} > {new_value}")
                            old_for_this_planet[f"{k}{s}"] = new_for_this_planet[
                                f"{k}{s}"
                            ]
                            keys_to_display.append(f"{k}{s}")
                        except (IndexError, KeyError):
                            pass
                    if verbose:
                        change_summary = vstack(
                            [
                                old_for_display[keys_to_display],
                                new_for_this_planet[keys_to_display],
                            ]
                        )
                        change_summary.add_column(
                            ["old", "new"], index=0, name="version"
                        )
                        display(change_summary)

                nothing_happened = False

        if nothing_happened:
            warnings.warn(warning_message)



from .setup_tests import *

from exoatlas import *


def test_planet_calculations():
    """
    Do the planetary-related population calculations work?
    """
    calculations = """
semimajoraxis_from_period,
semimajoraxis,
scaled_semimajoraxis,
eccentricity,
argument_of_periastron,
transit_impact_parameter,
insolation,
relative_insolation,
log_relative_insolation,
relative_cumulative_xuv_insolation,
teq,
planet_luminosity,
transit_depth_from_radii,
scaled_radius_from_radii,
scaled_radius,
transit_duration_from_orbit,
transit_duration,
surface_gravity,
density,
escape_velocity,
orbital_velocity,
impact_velocity,
escape_parameter,
scale_height
""".strip().split(
        ",\n"
    )
    test_calculations(calculations=calculations)


def test_stellar_calculations():
    """
    Do the stellar-related population calculations work?
    """
    calculations = """
stellar_luminosity_from_radius_and_teff,
stellar_luminosity,
distance_modulus
""".strip().split(
        ",\n"
    )
    test_calculations(calculations=calculations)


def test_observability_calculations():
    """
    Do the observability-related population calculations work?
    """
    calculations = """
angular_separation,
transmission_signal,
emission_signal,
reflection_signal,
stellar_brightness,
stellar_brightness_in_telescope_units,
depth_uncertainty,
depth_snr,
emission_snr,
reflection_snr,
transmission_snr
""".strip().split(
        ",\n"
    )
    test_calculations(calculations=calculations)


def test_calculations(calculations=[]):
    e = TransitingExoplanets()[::100]
    s = SolarSystem()
    for k in calculations:
        for p in [e, s]:
            print(f'trying to calculate "{k}" for {p}')
            p.get(k)
            p.get_uncertainty(k)


def test_add_columns():
    s = SolarSystem()
    new_column_name = "is_inhabited"
    new_column_data = (s.name() == "Earth") * 1
    new_column_uncertainty = (s.name() != "Earth") * 0.01
    s.add_column(
        name=new_column_name, data=new_column_data, uncertainty=new_column_uncertainty
    )
    s.is_inhabited()
    s.is_inhabited_uncertainty()


def test_add_calculations():
    s = SolarSystem()

    def f(self, distribution=False):
        """
        Surface Area (m)

        Calculate the surface area of a planet,
        based on its radius.
        """
        return 4 * np.pi * self.radius(distribution=distribution)

    s.add_calculation(name="surface_area", function=f)
    s.surface_area()
    s.surface_area_uncertainty()



from ..imports import *

__all__ = ["plot_seager", "plot_both_seager"]


# plot the mass radius models
def plot_seager(which="rock", **kw):
    """
    Plot a single model from Seager et al. (2007). These were
    originally point-clicked out of the GJ1214b discovery paper.

    Parameters
    ----------
    which : str
        Options are 'rock' or 'HHe'
    """

    assert (which.lower() == "rock") or (which.lower() == "hhe")

    # load the values
    file = os.path.join(code_directory, f"models/data/seager/gj1214mr_{which}.txt")
    m, r = np.transpose(np.loadtxt(file))

    # fit a simply polynomial to smooth out the messiness
    fit = np.poly1d(np.polyfit(np.log(m), np.log(r), 5))

    # plot over a grid of masses
    mm = np.logspace(np.log10(min(m)), np.log10(max(m)), 1000)
    return plt.plot(mm, np.exp(fit(np.log(mm))), **kw)


def plot_both_seager(**kw):
    """
    Plot both the rock and the cold H/He models from Seager et al. (2007)
    """
    zkw = dict(linewidth=4, alpha=0.25, zorder=1.0)
    zkw.update(**kw)
    plot_seager("hhe", label="cold H/He", color="darkorange", **zkw)
    plot_seager("rock", label="rock", color="black", **zkw)



from ...imports import *
from .plottable import *
from ...telescopes import *


class Flux(Plottable):
    source = "relative_insolation"
    label = "Planet Bolometric Flux Received\n(relative to Earth)"
    scale = "log"
    lim = [6e4, 2e-4]


class Teq(Plottable):
    source = "teq"
    scale = "log"
    lim = [200, 2000]

    def _update_label(self):

        self.label = (
            "Planet Equilibrum Temperature (K, f={f}, {albedo:.0%} albedo)".format(
                **self.kw
            )
        )


class CumulativeXUVFlux(Flux):
    source = "relative_cumulative_xuv_insolation"
    label = "Time-Integrated XUV Flux Received\n(relative to Earth)"


class ImpactVelocity(Plottable):
    source = "impact_velocity"
    label = "Estimated Impact Velocity (km/s)"
    scale = "log"


class Radius(Plottable):
    source = "radius"
    label = "Planet Radius (Earth radii)"
    scale = "log"
    lim = [0.3, 30]


class Mass(Plottable):
    source = "mass"
    label = "Planet Mass\n(Earth masses)"
    scale = "log"
    lim = [0.03, 3000]


class SemimajorAxis(Plottable):
    source = "semimajoraxis"
    label = "Semimajor Axis (AU)\n"
    scale = "log"
    lim = [0.001, 1000]


class AngularSeparation(Plottable):
    source = "angular_separation"
    label = "Angular Separation (arcsec)\n"
    scale = "log"
    lim = [0.001, 10]


class Contrast(Plottable):
    source = "imaging_contrast"
    label = "Planet-to-Star Contrast"
    scale = "log"
    lim = [1e-10, 1e-3]


class KludgedMass(Plottable):
    source = "kludge_mass"
    label = "Planet Mass or msini\n(Earth masses)"
    scale = "log"
    lim = [0.03, 4131]


class StellarTeff(Plottable):
    source = "stellar_teff"
    label = "Stellar Temperature (K)"
    scale = "linear"
    lim = [2000, 7000]


class StellarLuminosity(Plottable):
    source = "stellar_luminosity"
    label = "Stellar Luminosity (L_$\odot$)"
    scale = "log"
    lim = [None, None]


class Distance(Plottable):
    source = "distance"
    label = "Distance\n(parsecs)"
    scale = "log"
    lim = [5, 1000]


class Distance(Plottable):
    source = "distance"
    label = "Distance\n(parsecs)"
    scale = "log"
    lim = [5, 1000]


class EscapeVelocity(Plottable):
    source = "escape_velocity"
    label = "Escape\nVelocity\n(km/s)"
    scale = "log"
    lim = [2, 500]


class EscapeParameter(Plottable):
    source = "escape_parameter"
    label = "$\lambda = E_{grav}/E_{thermal}$"
    scale = "log"
    lim = [None, None]


class Density(Plottable):
    source = "density"
    label = "Planet Density\n(g/cm$^3$)"
    scale = "log"
    lim = [0.01, 100]


class StellarRadius(Plottable):
    source = "stellar_radius"
    label = "Stellar Radius\n(solar radii)"
    scale = "linear"
    lim = [0.0, 2.0]


class Period(Plottable):
    source = "period"
    label = "Period (days)\n"
    scale = "log"
    lim = [0.15, 365]


class Gmag(Plottable):
    source = "magnitude_gaia"
    label = "G (magnitude)\n"
    scale = "linear"
    lim = [3.5, 14.5]


class Depth(Plottable):
    source = "transit_depth"
    label = "Transit Depth\nof Planet"
    scale = "log"
    lim = [2e-6, 2e-1]


class StellarBrightness(Plottable):
    source = "stellar_brightness"
    scale = "log"
    lim = [1e2, 1e8]
    unit = u.Unit("ph s^-1 m^-2 micron^-1")

    def _update_label(self):
        self.label = f'Stellar Brightness at Earth at $\lambda={self.kw['wavelength'].to("micron").value:.1f}\mu$m\n({self.unit.to_string("latex_inline")})'


class StellarBrightnessTelescope(Plottable):
    source = "stellar_brightness_in_telescope_units"
    scale = "log"
    lim = [1e-3, 1e3]

    def _update_label(self):
        self.label = (
            f"Stellar Brightness at Earth at $\lambda={w}\mu$m\n({self.unit_string})"
        )

    def __init__(self, telescope_name="JWST", **kw):
        """
        Initialize the StellarBrightness plottable.
        It depends on wavelength, due to the thermal
        emission Planck spectrum.

        Parameters
        ----------
        telescope_name : None, str
            The telescope unit in which to express the
            stellar brightness. Options include:
            'JWST', 'Hubble', 'Kepler', 'TESS'

        wavelength : astropy.units.quantity.Quantity
            The wavelength at which we want the
            stellar brightness to be computed.

        R : float
            The spectral resolution at which the
            telescope will bin wavelengths..
            (Ignored if telescope is None.)

        dt : astropy.units.quantity.Quantity
            The time over which the telescope exposes.
            (Ignored if telescope is None.)
        """

        # do all the basic telescope + wavelength setup
        self.setup_telescope(telescope_name=telescope_name, **kw)

        # specify the unit to be used for plotting
        self.setup_unit()

        # initialize the basic plottable axis
        Plottable.__init__(self, **kw)

    def setup_telescope(self, telescope_name=None, **kw):
        """
        Setup the basics of the telescope, wavelength, and units.
        """

        # break if there's not telescope
        assert telescope_name is not None

        # keep track of the telescope (if any)
        self.telescope_name = telescope_name

        # define a unit of collecting power for the telescope
        self.telescope_unit = define_telescope_unit_by_name(self.telescope_name, **kw)

        # note the wavelength the telescope is set at
        self.wavelength = self.telescope_unit.wavelength
        self.R = self.telescope_unit.R

    def setup_unit(self):
        """
        Define the unit for plotting.
        """
        # what's the astropy unit associated with the axis?
        self.unit = lotsofphotons_unit / self.telescope_unit

        # what should go in the () in the axis label?
        self.unit_string = self.unit.to_string("latex_inline")

    def _update_label(self):
        """
        How should this plottable be labled on an axis?
        """
        # define the label, based on the wavelength and telescope
        w = self.wavelength.to(u.micron).value
        self.label = (
            f"Stellar Brightness at Earth at $\lambda={w}\mu$m\n({self.unit_string})"
        )


class DepthSNR(StellarBrightnessTelescope):
    source = "depth_snr"
    scale = "log"
    lim = [1e-3, 1e3]

    def setup_unit(self):
        # kludge to undo telescope brightness unit
        self.unit = u.Unit()
        self.unit_string = ""

    def _update_label(self):
        """
        How should this plottable be labled on an axis?
        """
        # define the label, based on the wavelength and telescope
        w = self.wavelength.to(u.micron).value
        self.label = f"S/N for Transit Depth\n{self.telescope_unit}"


class Transmission(Depth):
    source = "transmission_signal"

    def _update_label(self):
        mu = self.kw["mu"]
        self.label = f"Transit Depth\nof 1 Scale Height\n for $\mu$={mu} Atmosphere"


class TransmissionSNR(DepthSNR):
    scale = "log"
    size_normalization = 10
    source = "transmission_snr"

    def _update_label(self):
        mu = self.kw["mu"]
        w = self.wavelength.to(u.micron).value
        R = self.R
        self.label = f"S/N for Transit Depth\nof 1 Scale Height\n for $\mu$={mu} Atmosphere\n{self.telescope_unit}"


class Reflection(Depth):
    source = "reflection_signal"

    def _update_label(self):
        albedo = self.kw["albedo"]
        self.label = f"Reflected Light\nEclipse Depth\n({albedo:.0%} albedo)"


class ReflectionSNR(DepthSNR):
    scale = "log"
    size_normalization = 10
    source = "reflection_snr"

    def _update_label(self):
        albedo = self.kw["albedo"]
        w = self.wavelength.to(u.micron).value
        R = self.R
        self.label = f"S/N for Reflected Light\nEclipse Depth\n({albedo:.0%} albedo)\n{self.telescope_unit}"


class Emission(Depth):
    source = "emission_signal"

    def _update_label(self):
        albedo = self.kw["albedo"]
        f = self.kw["f"]
        w = self.kw["wavelength"].to_value("micron")
        self.label = f"Thermal Eclipse Depth\nat $\lambda={w}\mu m$\n(f={f:.2f}, {albedo:.0%} albedo)"


class EmissionSNR(DepthSNR):
    scale = "log"
    size_normalization = 10
    source = "emission_snr"

    def _update_label(self):
        albedo = self.kw["albedo"]
        f = self.kw["f"]
        w = self.wavelength.to(u.micron).value
        R = self.R
        self.label = f"S/N for Thermal\nEclipse Depth\n( f={f:.2f}, {albedo:.0%} albedo)\n{self.telescope_unit}"


class RightAscension(Plottable):
    source = "ra"
    label = "Right Ascension (hours)"
    scale = "linear"
    lim = [24, 0]
    unit = u.hourangle


class Declination(Plottable):
    source = "dec"
    label = "Declination (˚)"
    scale = "linear"
    lim = [-90, 90]
    unit = u.deg


preset_plottables = {}
local_variables = dict(**locals())
for k, v in local_variables.items():
    try:
        assert issubclass(v, Plottable)
        preset_plottables[k] = v
    except (AssertionError, TypeError):
        continue
preset_plottables.pop("Plottable")



"""
Define tool for downloading data from the NASA Exoplanet Archive.
"""

from ..downloader import *


class ExoplanetArchiveDownloader(Downloader):
    """
    ... to download Planetary Systems tables from the NASA Exoplanet Archive
    """

    # class-specific URL to access the archive API
    base = "https://exoplanetarchive.ipac.caltech.edu/TAP/sync?"

    # class-specific keywords to aid reading raw downloaded files into tables
    read_kw = dict(
        format="ascii", delimiter=",", fill_values=[("", np.nan), ("--", np.nan)]
    )

    # class-specific names of supported tables
    supported_tables = ["ps", "pscomppars"]

    def __init__(self, which_table="ps", format="csv", N=np.inf, planet="", **kw):
        """
        Initialize an Exoplanet Archive downloader.

        Parameters
        ----------
        which_table : str
            The name of the table to be downloaded.
            - 'ps' is Planetary Systems with one line per reference
            - 'pscomppars' is Planetary Systems Composite Parameters with one line per planet
        format : str
            The format of the table to download.
        N : int
            The maximum number of rows to return.
            This is mostly used for testing the downloader with small datasets.
        planet : str
            The name of planet to query, if desiring only a single planet.
            This is mostly used for testing the downloader with small datasets.
        **kw : dict
            Other keywords (expiration, timeout) are passed to the Downloader.
        """

        # keep track of what should be downloaded
        assert which_table in self.supported_tables
        self.which_table = which_table
        self.format = format
        self.N = N
        self.planet = planet
        Downloader.__init__(self, **kw)

    @property
    def top_string(self):
        # limit to a certain number of rows
        if (self.N == None) or (np.isfinite(self.N) == False):
            return ""
        else:
            return f"top+{self.N}+"

    @property
    def where_string(self):
        if (self.planet is "") or (self.planet is None):
            return ""
        else:
            return f"where+pl_name+like+'{self.planet}'".replace(" ", "%20")

    @property
    def from_string(self):
        return f"from+{self.which_table}+"

    @property
    def url(self):
        """
        Define the download URL to acces the online table.
        """

        # create the URL by stitching together keyword pairs
        url = (
            f"{self.base}"
            f"query=select+{self.top_string}*+"
            f"{self.from_string}"
            f"{self.where_string}"
            f"&format={self.format}"
        )

        return url

    @property
    def path(self):
        """
        Define a file path where the local copy of this file should be stored.
        """
        label = f"{self.top_string}{self.from_string}{self.where_string}".strip("+")
        return os.path.join(directories["data"], f"nasa-exoplanet-archive-{label}.txt")



from .setup_tests import *

from exoatlas.imports import *
from exoatlas.populations import *


def test_population():
    """
    Can we make a population from scratch from a table?
    """

    fake = QTable({x: [0] * 3 for x in core_descriptions}, masked=True)
    p = Population(standard=fake, label="fake")
    p._validate_columns()
    p.print_column_summary()


def test_solarsystem():
    """
    Can we make a population of Solar System planets?
    """
    for p in [
        SolarSystem,
        SolarSystemDwarfPlanets,
        SolarSystemMoons,
        SolarSystemMinorPlanets,
    ]:
        p()._validate_columns()


def test_exoplanets():
    """
    Can we make a population of confirmed exoplanets?
    """
    p = Exoplanets(remake=False)
    p.load_individual_references()
    p._validate_columns()

    # check that the background reference population for Exoplanets filters too
    f = p[:10]
    for x in f.hostname():
        assert x in f.individual_references.hostname()
    for x in f.individual_references.hostname():
        assert x in f.hostname()


def test_transitingexoplanets():
    """
    Can we make a population of confirmed transiting exoplanets?
    """
    p = TransitingExoplanets(remake=False)
    p._validate_columns()


def test_subsets():
    """
    Can we make a population of confirmed Kepler planets?
    """
    for x in [Kepler, NonKepler, TESS, NonTESS, Space, Ground, GoodMass, BadMass]:
        p = x(remake=False)
        p._validate_columns()

    with pytest.raises(NotImplementedError):
        q = TransitingExoplanetsSubset()


def test_indexing():
    """
    Can we make a population of confirmed transiting exoplanets?
    """
    p = Exoplanets(remake=False)

    # try different subsets
    b = p[5]
    c = p[0:10]
    with np.errstate(invalid="ignore"):
        d = p[p.stellar_radius() < 1.0 * u.Rsun]
    e = p["GJ 1214b"]
    f = p[["GJ 1214b", "LHS 1140b", "GJ 1132b"]]
    g = p[p.discovery_facility() == "Kepler"]
    h = p["TRAPPIST-1b"]
    i = p["TRAPPIST-1"]
    j = e + h
    k = f - e
    l = p.create_subset_by_name("GJ1214b")
    m = p.create_subset_by_hostname("GJ1214")
    coordinates = SkyCoord(e.ra(), e.dec())
    n = p.create_subset_by_position(coordinates)


def test_table():
    """
    Can we make a custom astropy table out of a Population?
    """

    p = SolarSystem()
    p.create_table()


def test_attributes():
    """
    Test the interaction with attributes.
    """
    p = SolarSystem()

    print(p.color)
    p.alpha = 0.5
    assert p._plotkw["alpha"] == 0.5

    for k in basic_columns:
        getattr(p, k)


def test_transiting(planet="GJ1214b"):
    with mock.patch("builtins.input", return_value=""):
        t = TransitingExoplanets()
        p = t[planet]

    summarize_planet(p)

    p.transmission_signal(mu=5, threshold=3)
    p.transmission_snr(
        mu=5, threshold=3, telescope_name="JWST", wavelength=3 * u.micron
    )

    p.emission_signal(wavelength=5 * u.micron)
    p.emission_snr(wavelength=5 * u.micron, telescope_name="JWST")

    p.reflection_signal(albedo=0.5)
    p.reflection_snr(albedo=0.5, wavelength=5 * u.micron, telescope_name="JWST")


"""def test_choosing():
    # this could be more informative
    e = TransitingExoplanets()[:10]
    e.semimajoraxis(visualize=True)
    e.semimajoraxis(distribution=True, visualize=True)
    e.semimajoraxis_uncertainty()"""



"""
Define tool for downloading data from the NASA Exoplanet Archive.
"""

from ..downloader import *
import requests, json


class SBDBDownloader(Downloader):
    """
    ... to download SBDB data from JPL Solar System Dynamics.
    """

    # class-specific URL to access the archive API
    base = "https://ssd-api.jpl.nasa.gov/sbdb_query.api"

    def __init__(self, minimum_diameter=100 * u.km, **kw):
        """
        Initialize a Solar System small body downloader.

        Parameters
        ----------
        **kw : dict
            Other keywords (expiration, timeout) are passed to the Downloader.
        """

        # keep track of what should be downloaded
        self.minimum_diameter = minimum_diameter
        Downloader.__init__(self, **kw)

    def download_fresh(self):
        """
        Download a brand new table from the JPL/SBDB Query API.
        """

        print(
            f"""
        Attempting to freshly download data from
        {self.base}
        
        This may take a long time, and we're dreadully 
        sorry that we haven't coded in a clever way to 
        tell you how long. Archive tables might be up to
        a few 100MB, so however long it takes your computer
        to download that is probably a reasonable-ish amount 
        of time to wait before getting stressed out. 
        """
        )

        self._parameters = {}
        self._parameters["fields"] = (
            "full_name,name,class,e,a,q,i,om,w,ma,tp,per,H,G,diameter,GM,density,rot_per,albedo"
        )
        self._parameters["sb-cdata"] = (
            '{"AND":["diameter|GT|' + f"{self.minimum_diameter.to_value('km')}" + '"]}'
        )

        # download the file
        self._downloaded = requests.get(
            self.base, params=self._parameters, timeout=self.timeout.to_value("second")
        )
        self.url = self._downloaded.url
        print("Download successful. Processing into a table.")

        # process into an astropy Table
        d = json.loads(self._downloaded.text)
        self._raw_data = d
        # create a dictionary to organize data
        organized = {}
        for i, k in enumerate(d["fields"]):
            this_column = [x[i] for x in d["data"]]
            if k in ["full_name", "name", "class"]:
                organized[k] = np.array(this_column).astype(str)
            else:
                organized[k] = np.array(this_column).astype(float)

        self.table = Table(organized)
        self.table.meta["url"] = self.url
        self.table.meta["minimum_diameter"] = self.minimum_diameter
        self.table.write(self.path, overwrite=True)

        print(f"Download successful! Saved file to {self.path}")

    @property
    def path(self):
        """
        Define a file path where the local copy of this file should be stored.
        """
        label = f"{self.minimum_diameter}km"
        return os.path.join(directories["data"], f"jpl-ssd-sbdb-{label}.hdf5")



from .Gallery import *
from .GridGallery import *


class PlanetGallery(Gallery):
    """
    This Gallery is designed to give a quick
    look at populations of transiting exoplanets,
    focusing on some of the physical attributes
    that strongly influence how planets work.
    """

    def setup_maps(self):
        """
        Create the figure + axes, and
        link the axes to Map objects.
        """
        # create the subplots
        self.fi, self.ax = self.create_subplots(
            nrows=2, ncols=3, width_ratios=[2, 4, 1], height_ratios=[1, 3]
        )

        # attach maps to each ax
        self.maps = {}
        self.maps["flux_x_radius"] = Flux_x_Radius(ax=self.ax[1, 1])
        # ErrorMap( xaxis=Flux, yaxis=Radius, ax=self.ax[1, 1])  #
        self.maps["mass_x_radius"] = Mass_x_Radius(ax=self.ax[1, 0])
        self.maps["flux_x_escape"] = Flux_x_EscapeVelocity(ax=self.ax[0, 1])
        # ErrorMap(xaxis=Flux, yaxis=EscapeVelocity, ax=self.ax[0, 1])
        self.maps["radius_x_radius"] = StellarRadius_x_PlanetRadius(ax=self.ax[1, 2])
        # ErrorMap(xaxis=StellarRadius, yaxis=Radius, ax=self.ax[1, 2])  #
        self.maps["mass_x_escape"] = Mass_x_EscapeVelocity(ax=self.ax[0, 0])
        # ErrorMap(xaxis=Mass, yaxis=EscapeVelocity, ax=self.ax[0, 0])

    def refine_maps(self):
        """
        Make small changes to Maps, after data are plotted.
        """

        #
        plt.sca(self.maps["flux_x_radius"].ax)
        self.maps["flux_x_radius"].plot_hz()
        self.maps["flux_x_radius"].ticks_simplify_exponents("y")
        self.maps["flux_x_radius"].add_teqaxis()
        self.maps["flux_x_radius"].remove_ylabel()
        self.maps["flux_x_radius"].add_legend(frameon=False, fontsize=8)

        plt.sca(self.maps["mass_x_radius"].ax)
        self.maps["mass_x_radius"].plot_seager(zorder=1e9)
        # self.maps['mass-radius'].ticks_enforce_multiple_oom("x")
        # self.maps['mass-radius'].ticks_simplify_exponents("xy")

        plt.sca(self.maps["mass_x_escape"].ax)
        self.maps["mass_x_escape"].remove_xlabel()

        plt.sca(self.maps["flux_x_escape"].ax)
        self.maps["flux_x_escape"].remove_xlabel()
        self.maps["flux_x_escape"].remove_ylabel()
        self.maps["flux_x_escape"].plot_constant_lambda()

        plt.sca(self.maps["radius_x_radius"].ax)
        self.maps["radius_x_radius"].remove_ylabel()


def add_size_explainer(ax=None, label="", x=0.95, y=0.95, ha="right", va="top", **kw):
    try:
        plt.sca(ax)
    except AttributeError:
        ax = plt.gca()

    plt.text(
        x, y, f"area $\propto$ {label}", ha=ha, va=va, transform=ax.transAxes, **kw
    )


class SimplePlanetGallery(Gallery):
    """
    This Gallery is designed to give a quick
    look at populations of transiting exoplanets,
    focusing on some of the physical attributes
    that strongly influence how planets work.
    """

    def setup_maps(self, figsize=(10, 5), **kw):
        """
        Create the figure + axes, and
        link the axes to Map objects.
        """
        # create the subplots
        self.fi, self.ax = self.create_subplots(
            nrows=1,
            ncols=3,
            width_ratios=[2, 4, 1],
            figsize=figsize,
            **kw,
        )

        # attach maps to each ax
        self.maps = {}
        self.maps["flux_x_radius"] = Flux_x_Radius(ax=self.ax[1])
        self.maps["mass_x_radius"] = Mass_x_Radius(ax=self.ax[0])
        self.maps["radius_x_radius"] = StellarRadius_x_PlanetRadius(ax=self.ax[2])

    def refine_maps(self):
        """
        Make small changes to Maps, after data are plotted.
        """

        #
        plt.sca(self.maps["flux_x_radius"].ax)
        plt.xlim(5e4, 2e-5)

        self.maps["flux_x_radius"].plot_hz()
        self.maps["flux_x_radius"].ticks_simplify_exponents("y")
        self.maps["flux_x_radius"].add_teqaxis()
        self.maps["flux_x_radius"].remove_ylabel()
        self.maps["flux_x_radius"].add_legend(
            frameon=False, fontsize=8, loc="lower right"
        )

        plt.sca(self.maps["mass_x_radius"].ax)
        self.maps["mass_x_radius"].plot_seager(zorder=1e9)
        # self.maps['mass-radius'].ticks_enforce_multiple_oom("x")
        # self.maps['mass-radius'].ticks_simplify_exponents("xy")

        plt.sca(self.maps["radius_x_radius"].ax)
        self.maps["radius_x_radius"].remove_ylabel()


class ObservableGallery(GridGallery):
    def __init__(self, wavelength=5 * u.micron, **kw):
        GridGallery.__init__(
            self,
            rows=[
                Radius(),
                StellarBrightness(wavelength=wavelength, lim=[0.5e4, 2e9]),
                Depth(),
                Transmission(),
                Emission(wavelength=wavelength),
                Declination(),
            ],
            cols=[Flux(lim=[0.5e4, 2e-2]), Distance(lim=[5, 2000]), RightAscension()],
            wavelength=wavelength,
            **kw,
        )

    def setup_maps(self, *args, telescope_name="JWST", wavelength=5 * u.micron, **kw):
        GridGallery.setup_maps(self, *args, **kw)

        snr_kw = dict(lim=[0, 1e2], scale="linear")
        for k, m in self.maps.items():
            if "depth" in k:
                m.plottable["size"] = DepthSNR(
                    telescope_name=telescope_name, wavelength=wavelength, **snr_kw, **kw
                )
            if "transmission" in k:
                m.plottable["size"] = TransmissionSNR(
                    telescope_name=telescope_name, wavelength=wavelength, **snr_kw, **kw
                )
            if "emission" in k:
                m.plottable["size"] = EmissionSNR(
                    telescope_name=telescope_name, wavelength=wavelength, **snr_kw, **kw
                )
            m.plottable["color"] = Flux(scale="log", lim=[1, 1e4])
            m.size_normalization = 6

    def refine_maps(self):
        GridGallery.refine_maps(self)
        self.maps["ra_x_dec"].ax.set_xticks([18, 12, 6, 0])
        RA_x_Dec.add_monthaxis(self, ax=self.maps["ra_x_dec"].ax, position=40)

        kw = dict(fontsize=6)
        m = self.maps["relative_insolation_x_transit_depth"]
        add_size_explainer(ax=m.ax, label=m.plottable["size"].label, **kw)

        m = self.maps["relative_insolation_x_transmission_signal"]
        add_size_explainer(ax=m.ax, label=m.plottable["size"].label, **kw)

        m = self.maps["relative_insolation_x_emission_signal"]
        add_size_explainer(ax=m.ax, label=m.plottable["size"].label, **kw)


class EverythingGallery(GridGallery):
    def __init__(self):
        GridGallery.__init__(
            self,
            cols=[Flux(lim=[5e4, 2e-5])],
            rows=[
                Radius(lim=[0.01, 30]),
                Mass(lim=[5e-5, 4100]),
                EscapeVelocity(lim=[0.01, 1e3]),
                Density,
                Depth(lim=[1e-8, 0.2]),
                Transmission(lim=[1e-8, 0.2]),
                Emission(lim=[1e-8, 0.2], wavelength=5 * u.micron),
                Distance(lim=[1, 2e3]),
            ],
            # StellarBrightness(telescope_name='JWST', wavelength=1*u.micron)],
            map_type=ErrorMap,
            mapsize=(7, 2),
        )

    def refine_maps(self):
        GridGallery.refine_maps(self)
        self.maps["relative_insolation_x_radius"].add_legend(fontsize=5)


'''class ObservableGallery(BuildablePlot):
    def plot(
        self,
        pops,
        observables=["depth", "emission", "transmission"],
        note="",
        telescope_name="JWST",
        per_transit=False,
        **wavelengthkw
    ):
        gs = self.create_gridspec(
            2,
            len(observables) + 2,
            wspace=0.05,
            hspace=0.05,
            bottom=0.21,
            top=0.98,
            right=0.98,
            left=0.08,
            figsize=(10, 8),
        )

        if note == "trim":
            for k, p in pops.items():
                pops[k] = p[p.radius < 2 * u.Rearth]

        kw = dict()
        if note != "bw":
            kw["color"] = "log_relative_insolation"
            kw["vmin"] = -1
            kw["vmax"] = 5
            for k, x in pops.items():
                if k == "solarsystem":
                    x.alpha = 1
                else:
                    x.alpha = 0.5

        hidealpha = 0.05
        showalpha = 1.0

        """if 'highlight' in note:
            for k in pops:
                if k in note:
                    pops[k].alpha = showalpha
                else:
                    pops[k].alpha = hidealpha

        if 'toi' in note:
            for k in pops:
                pops[k].alpha = hidealpha
            pops['toi'].alpha = showalpha
        else:
            pops['toi'].alpha = 0.

        if 'all-together' in note:
            pops['toi'].alpha=0.5
            pops['tess'].alpha=0.5
            pops['nontess'].alpha=0.5
        """

        column = 0

        if True:
            pr = Distance_x_Radius(**kw, **wavelengthkw)
            pr.build(pops=pops, ax=plt.subplot(gs[1, column]))

            db = Distance_x_Brightness(
                telescope_name=telescope_name, **kw, **wavelengthkw
            )
            db.build(pops=pops, ax=plt.subplot(gs[0, column]))
            db.remove_xlabel()
            column += 1

        if "depth" in observables:
            dr = Depth_x_Radius(
                size=DepthSNR(telescope_name=telescope_name, per_transit=per_transit),
                **kw,
                **wavelengthkw
            )
            dr.build(pops=pops, ax=plt.subplot(gs[1, column]))
            dr.remove_ylabel()

            x = Depth_x_Brightness(
                telescope_name=telescope_name,
                size=DepthSNR(telescope_name=telescope_name, per_transit=per_transit),
                **kw,
                **wavelengthkw
            )
            x.build(pops=pops, ax=plt.subplot(gs[0, column]))
            x.plot_sigma()
            x.remove_xlabel()
            x.remove_ylabel()
            column += 1

        if "emission" in observables:
            er = Emission_x_Radius(
                size=EmissionSNR(
                    telescope_name=telescope_name, per_transit=per_transit
                ),
                **kw,
                **wavelengthkw
            )
            er.build(pops=pops, ax=plt.subplot(gs[1, column]))
            er.remove_ylabel()

            x = Emission_x_Brightness(
                telescope_name=telescope_name,
                size=EmissionSNR(
                    telescope_name=telescope_name, per_transit=per_transit
                ),
                **kw,
                **wavelengthkw
            )
            x.build(pops=pops, ax=plt.subplot(gs[0, column]))
            x.plot_sigma()
            x.remove_xlabel()
            x.remove_ylabel()
            column += 1

        if "reflection" in observables:
            rr = Reflection_x_Radius(
                size=ReflectionSNR(
                    telescope_name=telescope_name, per_transit=per_transit
                ),
                **kw,
                **wavelengthkw
            )
            rr.build(pops=pops, ax=plt.subplot(gs[1, column]))
            rr.remove_ylabel()

            x = Reflection_x_Brightness(
                telescope_name=telescope_name,
                size=ReflectionSNR(
                    telescope_name=telescope_name, per_transit=per_transit
                ),
                **kw,
                **wavelengthkw
            )
            x.build(pops=pops, ax=plt.subplot(gs[0, column]))
            x.plot_sigma()
            x.remove_xlabel()
            x.remove_ylabel()
            column += 1

        if "transmission" in observables:
            tr = Transmission_x_Radius(
                size=TransmissionSNR(
                    telescope_name=telescope_name, per_transit=per_transit
                ),
                **kw,
                **wavelengthkw
            )
            tr.build(pops=pops, ax=plt.subplot(gs[1, column]))
            tr.remove_ylabel()

            x = Transmission_x_Brightness(
                telescope_name=telescope_name,
                size=TransmissionSNR(
                    telescope_name=telescope_name, per_transit=per_transit
                ),
                **kw,
                **wavelengthkw
            )
            x.build(pops=pops, ax=plt.subplot(gs[0, column]))
            x.plot_sigma()
            x.remove_xlabel()
            x.remove_ylabel()
            column += 1

        if True:
            fr = Flux_x_Radius(**kw)
            fr.xlabel = "Bolometric Flux Received\n(relative to Earth)"
            fr.build(pops=pops, ax=plt.subplot(gs[1, column]))
            fr.remove_ylabel()
            column += 1
'''



from ...imports import *
import astroplan as ap
from astroplan.plots import plot_airmass

sbo = ap.Observer(
    longitude=-105.2630 * u.deg,
    latitude=40.00371 * u.deg,
    elevation=1653 * u.m,
    name="Sommers-Bausch Observatory",
    timezone="US/Mountain",
)


def altaz(self, where=sbo, when="tonight", **kw):
    """
    Altitude, Azimuth (astropy.coordinates.SkyCoord)

    This returns an `astropy.coordinates.SkyCoord` object
    representing the altitude and azimuth of the objects
    in the population from a particular observer's
    location on Earth (`where`) at a particular moment (`when`).

    Caveats:
    It's (currently) meaningless for Solar System objects.
    It doesn't propagate uncertainties.
    It doesn't include proper motions or parallax.

    Parameters
    ----------
    where : astroplan.Observer
        This must be an astroplan Observer object
        encoding the location (and timezone) of an
        Earth-bound observatory.
    when : str, astropy.time.Time
        This can be a Time object specifying the moment at
        which the altitude and azimuth should be calculated.
        Or, it can be the string "tonight" meaning to use
        the closest local midnight to the current time.
    """

    if when == "tonight":
        when = where.midnight(Time.now())
    coordinates = SkyCoord(ra=self.ra(), dec=self.dec())
    altaz = where.altaz(when, coordinates)
    return altaz


def airmass(self, where=sbo, when="tonight", **kw):
    """
    Airmass (unitless)

    Calculate the airmass of a target from a particular observer's
    location on Earth (`where`) at a particular moment (`when`).

    Parameters
    ----------
    where : astroplan.Observer
        This must be an astroplan Observer object
        encoding the location (and timezone) of an
        Earth-bound observatory.
    when : str, astropy.time.Time
        This can be a Time object specifying the moment at
        which the altitude and azimuth should be calculated.
        Or, it can be the string "tonight" meaning to use
        the closest local midnight to the current time.
    """

    secz = self.altaz(where=where, when=when).secz

    airmass = np.where(secz > 0, secz, np.nan)
    return airmass


def plot_airmass_for_transit(row, max_airmass=2.0, savefig=False):
    """
    Make an airmass plot for one transit.

    For each row of a table produced by `show_upcoming_transits`,
    this function produces a plot showing the target's airmass
    from a particular location, with the transit indicated.

    Parameters
    -----------
    max_airmass : float
        The worst airmass to show on the plot.
    savefig : bool
        Should plots be saved out automatically
        into a folder called "airmass-plots"?
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        plt.figure()

        # use astroplan to do most of the work
        plot_airmass(
            targets=row["target"],
            observer=row.meta["where"],
            time=row["midpoint"],
            brightness_shading=True,
            altitude_yaxis=True,
            max_airmass=max_airmass,
            style_kwargs=dict(color="black"),
        )

        # plot the transit as vertical lines
        plt.axvline(
            row["midpoint"].plot_date,
            color="black",
            linestyle="--",
            label="mid-transit",
        )
        ingress_kw = dict(color="black", linestyle=":")
        plt.axvline(row["ingress"].plot_date, **ingress_kw)
        plt.axvline(row["egress"].plot_date, **ingress_kw)

        # add title
        plt.title(
            f'{row["name"]}\n{row.meta["where"].name} | {row["midpoint"].iso} UTC'
        )

        # set xlimit to be exactly sunset to sunrise
        plt.xlim(
            row.meta["where"].sun_set_time(row["midpoint"], which="previous").plot_date,
            row.meta["where"].sun_rise_time(row["midpoint"], which="next").plot_date,
        )

    # save the figure, if desired
    if savefig:
        mkdir("airmass-plots")
        plt.savefig(
            f"airmass-plots/{clean(row["name"])}-{row["midpoint"].iso.split()[0]}.png",
            bbox_inches="tight",
        )


def show_upcoming_transits(
    self,
    where=sbo,
    when="now",
    window=3 * u.day,
    allow_partial_transits=False,
    min_altitude=30 * u.deg,
    max_altitude=90 * u.deg,
    visualize=True,
):
    """
    Calculate all observable transits for a population.

    Parameters
    ----------
    where : astroplan.Observer
        This must be an astroplan Observer object
        encoding the location (and timezone) of an
        Earth-bound observatory.
    when : str, astropy.time.Time
        Specify the start of the window to start
        considering to search for transits, either
        as an astropy Time or as the string "now",
        meaning starting tonight for whenver you
        call the function.
    window : astropy.units.Quantity
        How long of a window should we consider for
        searching for potentially observable transits?
    allow_partial_transits : bool
        If True, only the transit midpoint must
        meet the constraints to be observable.
        If False, ingress + midpoint + egress must
        all together meet the constraints.
    min_altitude : astropy.units.Quantity
        The lowest altitude to consider observable.
        (30 degrees corresponds to airmass 2)
    max_altitude : astropy.units.Quantity
        The highest altitude to consider observable.
        (some telescopes can't observe at zenith!)
    visualize : bool
        Should we automatically produce an airmass
        plot for each observable transit event?
    """

    # construct skeleton table and a place to store transit events
    initial_planning_table = self.create_table(
        [
            "name",
            "ra",
            "dec",
            "magnitude_gaia",
            "transit_depth",
            "stellar_radius",
            "distance",
        ]
    )
    tables_for_individual_planets = []

    # define the start of the observing window
    if when == "now":
        when = Time.now()
    start_of_observing_window = where.midnight(when) - 0.5 * u.day

    # (keep track of if we need to warn about bad data)
    need_to_warn = True

    # there are some ~1s timing warnings we can probably ignore
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        # lopp through planets
        for i in range(len(self)):

            # set the target location on the sky
            target = ap.FixedTarget(
                coord=SkyCoord(ra=self.ra()[i], dec=self.dec()[i]), name=self.name()[i]
            )
            try:
                # set the orbital parameters
                es = ap.EclipsingSystem(
                    primary_eclipse_time=Time(self.transit_midpoint()[i], format="jd"),
                    orbital_period=self.period()[i],
                    duration=self.transit_duration()[i],
                )

                # calculate enough upcoming transits to fill dt
                n_transits = int(np.ceil(window / self.period()[i]))

                # (n_transits,) Time array
                midtransit_times = es.next_primary_eclipse_time(
                    start_of_observing_window, n_eclipses=n_transits
                )
                # (n_transits, 2) Time array
                ingress_and_egress_times = es.next_primary_ingress_egress_time(
                    when, n_eclipses=n_transits
                )
            except ValueError:
                # warn us if a planet has bad planning data
                if need_to_warn:
                    print("We could not predict transits for these planets:\n")
                    need_to_warn = False
                print(
                    f"☹️ {self.name()[i]:>20} | T0={self.transit_midpoint()[i]:>20} | P={self.period()[i]:10.3f} | dt={self.transit_duration()[i]:10.3f} ☹️"
                )
                continue

            # decide whether ingress/egress must be observable too
            if allow_partial_transits:
                times_ingress_egress = None
            else:
                times_ingress_egress = ingress_and_egress_times

            # calculate which of all possible events are observable
            ok = ap.is_event_observable(
                constraints=[
                    ap.AtNightConstraint(max_solar_altitude=-12 * u.deg),
                    ap.AltitudeConstraint(min=min_altitude, max=max_altitude),
                ],
                observer=where,
                target=target,
                times=midtransit_times,
                times_ingress_egress=times_ingress_egress,
            ).flatten()

            # if there are transits, record them!
            N = sum(ok)
            if N > 0:
                table_of_transits_for_this_planet = Table(
                    dict(
                        name=[self.name()[i]] * N,
                        ingress=ingress_and_egress_times[ok, 0],
                        midpoint=midtransit_times[ok],
                        egress=ingress_and_egress_times[ok, 1],
                        target=[target] * N,
                    )
                )
                tables_for_individual_planets.append(table_of_transits_for_this_planet)

    if len(tables_for_individual_planets) == 0:
        print("📭 No transits were found. Sorry! 📭")
        return None

    # stack and join individual planets into one giant table with metadata
    table_of_all_transits = vstack(tables_for_individual_planets)
    transit_planning_table = join(table_of_all_transits, initial_planning_table)
    transit_planning_table.meta["where"] = where
    transit_planning_table.meta["when"] = when
    transit_planning_table.meta["window"] = window

    if visualize:
        for row in transit_planning_table:
            plot_airmass_for_transit(row, savefig=True)

    return transit_planning_table



from .version import *

try:
    from IPython.display import display
except ImportError:
    display = print

# imports that are need by many exoatlas subsections
from ast import Import
import os, sys, time, shutil, warnings, copy, glob, inspect
from tqdm import tqdm

# (possibly different on Mac, Linux, Windows, even for Python versions 3.8-3.12)
try:
    from importlib.resources import files
except (ModuleNotFoundError, AttributeError, ImportError):
    from importlib_resources import files

code_directory = files(import_name)


import numpy as np, matplotlib.pyplot as plt, matplotlib.animation as animation
from matplotlib.ticker import ScalarFormatter, FormatStrFormatter, LogLocator


from astropy.utils.exceptions import AstropyDeprecationWarning

warnings.simplefilter("ignore", category=AstropyDeprecationWarning)

# this function downloads a file and returns its filepath
from astropy.utils.data import download_file
from astropy.io import ascii
from astropy.table import Table, QTable, Row, Column, vstack, join, setdiff
from astropy.uncertainty import Distribution
from astropy.visualization import quantity_support
from astropy.coordinates import SkyCoord

quantity_support()

# units and constants from astropy
import astropy.units as u, astropy.constants as con
from astropy.time import Time


def mkdir(path):
    """A mkdir that doesn't complain if it fails."""
    try:
        os.mkdir(path)
    except:
        pass


import matplotlib.colors as co


def name2color(name):
    """
    Return the 3-element RGB array of a given color name.

    Parameters
    ----------
    name : str
            The name ('black') or hex ('#000000')
            for a particular color.

    Returns
    -------
    rgb : tuple
            Three-elements array of floats,
            expressing brightness in RGB.
    """
    if "#" in name:
        h = name
    else:
        h = co.cnames[name].lower()
    return co.hex2color(h)


# create a directory structure ()
try:
    # search for an environment variable
    base = os.getenv("EXOATLAS_DATA")
    assert base is not None
except AssertionError:
    # otherwise put it in the local directory
    cwd = os.getcwd()
    base = os.path.join(cwd, "downloads-for-exoatlas")
mkdir(base)


def locate_local_data():
    print("💾 `exoatlas` archive data will be stored in:")
    print(base)


directories = dict(data=os.path.join(base, "data/"))
for k in directories.keys():
    mkdir(directories[k])


def reset_standardized_data():
    files = glob.glob(os.path.join(directories["data"], "standardized-*"))
    if "y" in input(f"Are you sure you want to wipe {files}? [y/N]"):
        for f in files:
            os.remove(f)


def reset_local_data():
    if "y" in input(
        "Are you sure you want to wipe all " "local exoatlas data files? [y/N]"
    ):
        shutil.rmtree(directories["data"])
        mkdir(directories["data"])
        print(f"Removed all local data from {directories['data']}")


# some kludge for dealing with Python 3 vs 2?
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError


def clean(s):
    """
    A wrapper function to clean up complicated strings.
    """
    bad = """ !@#$%^&*()+-_'",./<>?;\n\r\t"""
    cleaned = str(s) + ""
    for c in bad:
        cleaned = cleaned.replace(c, "")
    return cleaned


def time_from_modified(filename):
    """
    How long ago was this file last modified?
    """
    try:
        dt = Time.now().unix - os.path.getmtime(filename)
        return dt / 60 / 60 / 24
    except FileNotFoundError:
        return np.inf


def check_if_needs_updating(filename, maximum_age=1.0):
    """
    Do a (possibly interactive) check to see if this file
    is so old that it needs to be updated.

    Returns
    -------
    old : bool
            True if the file is so old it needs updating
            False if the file doesn't need updating
    """

    # how long ago was the data updated?
    dt = time_from_modified(filename)

    old = False
    if dt == np.inf:
        old = True
    elif dt > maximum_age:
        print(f"{filename} is {dt:.3f} days old.")
        old = "y" in input("Should it be updated? [y/N]").lower()

    return old


def one2another(bottom="white", top="red", alphabottom=1.0, alphatop=1.0, N=256):
    """
    Create a cmap that goes smoothly (linearly in RGBA) from "bottom" to "top".
    """
    try:
        rgb_bottom = bottom[0:3] + 0
    except:
        rgb_bottom = name2color(bottom or "white")
    try:
        rgb_top = top[0:3] + 0
    except:
        rgb_top = name2color(top or "black")

    r = np.linspace(rgb_bottom[0], rgb_top[0], N)
    g = np.linspace(rgb_bottom[1], rgb_top[1], N)
    b = np.linspace(rgb_bottom[2], rgb_top[2], N)
    a = np.linspace(alphabottom, alphatop, N)
    colors = np.transpose(np.vstack([r, g, b, a]))
    cmap = co.ListedColormap(colors, name="{bottom}2{top}".format(**locals()))
    return cmap


class AtlasError(ValueError):
    pass


# KLUDGE

import warnings

warnings.filterwarnings("ignore", message="invalid escape sequence")


def remove_unit(x):
    """
    Remove astropy units from a variable.

    A simple wrapper to help make sure that we're dealing
    just with numerical values, instead of astropy Quantitys
    with units.
    """

    if isinstance(x, u.quantity.Quantity):
        return x.value
    else:
        return x



from .setup_tests import *

from exoatlas import *


def test_airmass(visualize=False):
    e = TransitingExoplanets()

    positions = e.altaz()
    airmass = e.airmass()


def test_show_upcoming_transits(visualize=False):
    e = TransitingExoplanets()
    nearby = e[e.distance() < 15 * u.pc]
    transits = nearby.show_upcoming_transits(window=10 * u.day, visualize=visualize)



# FIXME! right now it's not obvious except to a pro user what items
# are attributes and which are methods which need to be called as
# functions. We should try to make that more transparent/easy!

core_basic_descriptions = {
    "name": "name of the planet",
    "hostname": "name of the host star",
}

core_stellar_descriptions = {
    "ra": "Right Ascension of the system",
    "dec": "Declination of the system",
    "distance": "distance to the system",
    "stellar_teff": "stellar effective temperature",
    "stellar_mass": "stellar mass",
    "stellar_radius": "stellar radius",
}


# transit_columns = [
#    "period",
#    "semimajoraxis",
#    "eccentricity",
#    "argument_of_periastron",
#    "inclination",
#    "transit_midpoint",
#    "transit_duration",
#    "transit_depth",
#    "stellar_teff",
#    "stellar_mass",
#    "stellar_radius",
#    "radius",
#    "mass",
#    "transit_scaled_semimajoraxis",
#    "transit_impact_parameter",
# ]

core_planet_descriptions = {
    "discovery_facility": "telescope/project that found this planet",
    "period": "orbital period of the planet",
    "eccentricity": "eccentricity",
    "argument_of_periastron": "argument of periastron",
    "radius": "planet radius",
    "mass": "planet mass",
}

core_transit_descriptions = {
    "transit_midpoint": "a transit midpoint",
    "transit_duration": "duration of the transit",
    "transit_depth": "fraction of starlight the planet blocks",
    "transit_impact_parameter": "(transit-derived) impact parameter b",
    "scaled_semimajoraxis": "(transit-derived) scaled orbital distance a/R*",
    "scaled_radius": "ratio of planet radius to stellar radius",
}

derived_stellar_descriptions = {
    "stellar_luminosity": "bolometric luminosity of the star",
    "distance_modulus": "apparent magnitude - absolute magnitude",
}
# calculated_columns = [
#    "scaled_semimajoraxis",
#    "b",
#    "insolation",
#    "relative_insolation",
#    "log_relative_insolation",
#    "teq",
#    "planet_luminosity",
#    "density",
#    "surface_gravity",
#    "distance_modulus",
#    "escape_velocity",
#    "escape_parameter",
#    "stellar_luminosity",
# ]

derived_planet_descriptions = {
    "inclination": "orbital inclination",
    "semimajoraxis": "the semimajor axis of the planet's orbit",
    "scaled_semimajoraxis": "scaled orbital distance a/R*",
    "density": "density of the planet",
    "insolation": "bolometric energy flux the planet receives from its star",
    "relative_insolation": "insolation relative to Earth",
    "log_relative_insolation": "log10(insolation relative to Earth)",
    "teq": "equilibrium temperature of the planet (assuming 0 albedo)",
    "planet_luminosity": "power emitted by the planet (assuming 0 albedo)",
    "surface_gravity": "surface gravitational acceleration of the planet",
    "scale_height": "scale height of an H2-rich atmosphere",
    "escape_velocity": "escape velocity of the planet",
    "escape_parameter": "ratio of gravitational potential to thermal energy for an H atom",
}

derived_observability_descriptions = {
    "transmission_signal": "transit depth of one scale height of atmosphere (a function of mean molecular weight)",
    "emission_signal": "thermal-emission eclipse depth (a function of wavelength)",
    "reflection_signal": "reflected-light eclipse depth (a function of albedo)",
    "depth_uncertainty": "predicted, photon-limited transit depth uncertainty",
    "transmission_snr": "S/N for transmission",
    "emission_snr": "S/N for emission",
    "reflection_snr": "S/N for reflection",
}

core_descriptions = (
    core_basic_descriptions
    | core_stellar_descriptions
    | core_planet_descriptions
    | core_transit_descriptions
)
derived_descriptions = (
    derived_stellar_descriptions
    | derived_planet_descriptions
    | derived_observability_descriptions
)
all_descriptions = core_descriptions | derived_descriptions

basic_columns = list(core_basic_descriptions.keys())
core_columns = list(core_descriptions.keys())
derived_columns = list(derived_descriptions.keys())
all_columns = list(all_descriptions.keys())


def describe_columns():
    """
    Describe some of the common columns you might want to
    access in an exoplanet population.
    """
    N = max([len(x) for x in all_descriptions])
    f = "{:>" + str(N) + "} = {}"
    for k, v in all_descriptions.items():
        print(f.format(k, v))



from ..imports import *
from matplotlib.colors import Normalize

__all__ = ["ink_errorbar"]


def ink_errorbar(
    x,
    y,
    yerr=None,
    xerr=None,
    c=None,
    cmap=one2another("white", "black"),
    vmin=0,
    vmax=1,
    alpha=1.0,
    **kw
):
    """
    Draw errorbars, allowing each to have its own color.

    This is meant to be a drop-in replacement for plt.errorbar that
    can more easily give different colors to each point, using the c=
    and cmap= keywords. It's a little slow because it needs to draw the
    individual errorbars one-by-one, but for datasets smaller than a
    few thousand elements, it shouldn't be too dreadful.

    Parameters
    ----------

    x : array
        The x values
    y : array
        The y values
    yerr : array
        The y uncertainty value
    xerr : array
        The x uncertainty value
    c : array
        The array of numerical values to convert to colors.
        By default, these should fall between 0 and 1.
    cmap : colormap
        A colormap to convert c= values to colors.
    vmin, vmax : float, float
        The minimumu and maximum values to normalize the cmap.
    alpha : float or array
        The transparency by which *all* points will be multiplied.
    **kw : dict
        All other keywords will be passed to plt.errorbar.

    """

    i_sorted = np.argsort(c)

    # calculate Nx4 array of colors
    norm = Normalize(vmin=vmin, vmax=vmax)
    colors = cmap(norm(c))

    colors[:, -1] *= alpha

    # loop over data points
    for i in i_sorted:  # tqdm()
        # pick this data point's color
        # color = np.array(colors[i]).astype(np.float)
        # assert(len(color) == 4)

        # make more transparent, if alpha is <1
        # try:
        #    color[-1] *= alpha[i]
        # except TypeError:
        #    color[-1] *= alpha

        # rgba = (color[0], color[1], color[2], color[3])

        # set the color of every part of the point to be plotted
        kw["color"] = colors[i, :]
        kw["ecolor"] = colors[i, :]
        # kw['markeredgecolor'] = colors[i,:]
        # kw['markerfacecolor'] = colors[i,:]

        # try:
        # assert(len(zorder) > 1)
        # assert(len(zorder) == len(x))
        #    kw['zorder'] = zorder[i]
        # except (TypeError, KeyError):
        #    kw['zorder'] = zorder

        # reshape the errors as needed (for asymmetric errors)
        if yerr is not None:
            if len(yerr.shape) == 1:
                yerrtoplot = yerr[i]
            if len(yerr.shape) == 2:
                yerrtoplot = yerr[:, i].reshape(2, 1)
        else:
            yerrtoplot = None

        if xerr is not None:
            if len(xerr.shape) == 1:
                xerrtoplot = xerr[i]
            if len(xerr.shape) == 2:
                xerrtoplot = xerr[:, i].reshape(2, 1)
        else:
            xerrtoplot = None

        # actually draw the error bar for this one
        # print(kw['zorder'])
        assert np.isfinite(x[i])
        assert np.isfinite(y[i])
        assert np.isfinite(xerrtoplot).all()
        assert np.isfinite(yerrtoplot).all()

        plt.errorbar(x[i], y[i], xerr=xerrtoplot, yerr=yerrtoplot, **kw)



from .setup_tests import *

from exoatlas.imports import *
from exoatlas.populations import *
from exoatlas.populations.pineda_skew import *


def test_skew(N_data=25):
    mu = np.zeros(N_data)
    sigma_lower = 10 ** np.linspace(-1, 1, N_data)
    sigma_upper = 10 ** np.linspace(1, -1, N_data)
    x = make_skew_samples_from_lowerupper(
        mu=mu, sigma_lower=sigma_lower, sigma_upper=sigma_upper
    )
    measured_lower, measured_upper = np.percentile(
        x,
        np.array([0.5 - gaussian_central_1sigma / 2, 0.5 + gaussian_central_1sigma / 2])
        * 100,
        axis=-1,
    )

    x_axis = np.log10(sigma_upper / sigma_lower)  # np.arange(N_data) #
    expected_kw = dict(color="gray", zorder=-1, alpha=0.3)
    fi, ax = plt.subplots(2, 1, sharex=True)
    plt.sca(ax[0])
    plt.plot(
        x_axis,
        mu,
        linestyle="-",
        label=r"requested $\mu^{+\sigma_{upper}}_{-\sigma_{lower}}$",
        **expected_kw
    )
    plt.plot(x_axis, mu - sigma_lower, linestyle="--", **expected_kw)
    plt.plot(x_axis, mu + sigma_upper, linestyle="--", **expected_kw)
    if N_data == 1:
        w = 1
    else:
        w = np.median(np.gradient(x_axis)) * 0.9
    plt.violinplot(
        x.T,
        positions=x_axis,
        widths=w,
        showextrema=False,
        quantiles=[
            [0.5 - gaussian_central_1sigma / 2, 0.5, 0.5 + gaussian_central_1sigma / 2]
        ]
        * N_data,
    )
    plt.legend(frameon=False)
    plt.ylabel("simulated distributions")

    plt.sca(ax[1])
    plt.plot(x_axis, (mu - measured_lower) / sigma_lower, marker="o", label="upper")
    plt.plot(x_axis, (measured_upper - mu) / sigma_upper, marker="o", label="lower")
    plt.legend(frameon=False)
    plt.xlabel(r"$\log_{10}(\sigma_{upper}/\sigma_{lower})$")
    plt.ylabel(r"$\sigma_{68\%, measured}/\sigma_{injected}$")


def test_make_astropy_distribution(key="radius"):
    e = TransitingExoplanets()
    mu = e.get_values_from_table(key)
    lower, upper = e.get_uncertainty_lowerupper_from_table(key)
    inject_kw = dict(color="gray", linewidth=5, alpha=0.2)
    fi, ax = plt.subplots(2, 1, sharex=True)
    plt.sca(ax[0])
    plt.plot(-lower, **inject_kw)
    plt.plot(upper, **inject_kw)
    plt.title(key)
    plt.ylabel("upper + lower\n" + r"$\sigma_{samples}$ and $\sigma_{table}$")
    d = e.get_values_from_table(key, distribution=True)
    sample_lower, sample_upper = d.pdf_percentiles([15.8, 84.2])
    sample_kw = dict(color="black")
    plt.plot(sample_lower - mu, **sample_kw)
    plt.plot(sample_upper - mu, **sample_kw)
    plt.sca(ax[1])
    plt.plot((sample_lower - mu) / lower, **sample_kw)
    plt.plot((sample_upper - mu) / upper, **sample_kw)
    plt.ylabel("upper + lower\n" + r"$\sigma_{samples}/\sigma_{table}$")


def test_uncertainties():
    """
    Can we estimate uncertainties for lots of quantities?
    """

    p = SolarSystem()

    uncertainty = p.get_uncertainty("radius")
    assert np.all(uncertainty == 0 * u.Rearth)

    upper, lower = p.get_uncertainty_lowerupper("radius")
    assert np.all(lower == 0 * u.Rearth)
    assert np.all(upper == 0 * u.Rearth)

    bad = p.get_uncertainty("distance")
    assert np.all(np.isnan(bad))


if __name__ == "__main__":  # pragma: no cover
    outputs = {k.split("_")[-1]: v() for k, v in locals().items() if "test_" in k}



from ...imports import *
from ...models.chen import *


def semimajoraxis_from_period(self, distribution=False, **kw):
    """
    Planet Semi-major Axis (a, AU)

    Calculate "a" from the period and stellar mass.
    This might be used if no semi-major axis is
    available in the standardized table.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    P = self.get("period", distribution=distribution)
    M = self.get("stellar_mass", distribution=distribution)
    G = con.G
    a = ((G * M * P**2 / 4 / np.pi**2) ** (1 / 3)).to("AU")
    return a


def semimajoraxis_from_transit_scaled_semimajoraxis(self, distribution=False, **kw):
    """
    Planet Semi-major Axis (a, AU)

    Calculate "a" from the transit-derived a/Rs.
    This might be used if no semi-major axis is
    available in the standardized table.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    a_over_R = self.get("transit_scaled_semimajoraxis", distribution=distribution)
    R = self.get("stellar_radius", distribution=distribution)
    a = (a_over_R * R).to("AU")
    return a


def semimajoraxis(self, distribution=False, **kw):
    """
    Planet Semi-major Axis (a, AU)

    Retrieve "a" first from the standardized table,
    then from period/mass, then from transit a/R.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    a = self._choose_calculation(
        methods=[
            "semimajoraxis_from_table",
            "semimajoraxis_from_period",
            "semimajoraxis_from_transit_scaled_semimajoraxis",
        ],
        distribution=distribution,
        **kw,
    )
    return a


def scaled_semimajoraxis_from_semimajoraxis(self, distribution=False, **kw):
    """
    Planet Scaled Semi-major Axis (a/R*, unitless)

    Calculate "a/Rs" from the semimajor axis a and
    radius Rs. This might be used if the table has
    no direct transit-derived value in its table.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    a = self.get("semimajoraxis", distribution=distribution)
    R = self.get("stellar_radius", distribution=distribution)
    return (a / R).decompose()


def scaled_semimajoraxis(self, distribution=False, **kw):
    """
    Planet Scaled Semi-major Axis (a/R*, unitless)

    Retrieve "a/Rs" first from the transit-derived value
    in the standardized table, then from a and Rs.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    a_over_Rs = self._choose_calculation(
        methods=[
            "transit_scaled_semimajoraxis_from_table",
            "scaled_semimajoraxis_from_semimajoraxis",
        ],
        distribution=distribution,
        **kw,
    )
    return a_over_Rs


def eccentricity(self, distribution=False, **kw):
    """
    Planet Orbital Eccentricity (e, unitless)

    The eccentricity of the planet's orbit. If eccentricity is
    `nan` in the standardized table, this method will quietly
    assume it to be zero.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """

    # pull out the actual values from the table
    e = self.get_values_from_table("eccentricity", distribution=distribution)

    # replace nans with 0
    if distribution == False:
        bad = np.isfinite(e) == False
        e[bad] = 0
    # (by not doing anything to replace nan values with 0 if `distribution==True`,
    # planets missing original values will end up with `nan` uncertainties)

    return e


def argument_of_periastron(self, distribution=False, **kw):
    """
    Planet Orbital Argument of Periastron ($\omega$, degrees)

    The argument of periastron of the planet's orbit, $\omega$.
    If it is `nan` (often because eccentricity is zero), then
    this will be assumed to be 0.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    # pull out the actual values from the table
    argument_of_periastron = self.standard["argument_of_periastron"].copy()

    # replace nans with 0
    if distribution == False:
        bad = np.isfinite(argument_of_periastron) == False
        argument_of_periastron[bad] = 0 * u.deg
    # (by not doing anything to replace nan values with 0 if `distribution==True`,
    # planets missing original values will end up with `nan` uncertainties)

    return argument_of_periastron


def transit_impact_parameter_from_inclination(self, distribution=False, **kw):
    """
    Planet Semi-major Axis (a, AU)

    Calculate "b" from the inclination, scaled semimajor axis,
    eccentricity, and argument of periastron. This might be
    used if no direct transit-derived impact parameter is
    available in the standardized table.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    # extract necessary quantities
    a_over_Rs = self.scaled_semimajoraxis(distribution=distribution)
    i = self.inclination(distribution=distribution)
    e = self.eccentricity(distribution=distribution)
    omega = self.argument_of_periastron(distribution=distribution)

    # calculate impact parameter based on instantaneous distance at transit
    b = a_over_Rs * np.cos(i) * ((1 - e**2) / (1 + e * np.sin(omega)))
    return b


def transit_impact_parameter(self, distribution=False, **kw):
    """
    Planet Impact Parameter (b)

    Retrieve "b" first from the transit-derived value
    in the standardized table, then from inclination.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    b = self._choose_calculation(
        methods=[
            "transit_impact_parameter_from_table",
            "transit_impact_parameter_from_inclination",
        ],
        distribution=distribution,
        **kw,
    )
    return b


# the 1360 W/m^2 that Earth receives from the Sun
earth_insolation = (1 * u.Lsun / 4 / np.pi / u.AU**2).to(u.W / u.m**2)


def insolation(self, distribution=False, **kw):
    """
    Planet Insolation (S, W/m**2)

    Calculate the insolation the planet receives from its star,
    given the luminosity of the star and the semimajor axis,
    expressed in units of W/m**2. (For reference, Earth
    receives 1360 W/m**2).

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """

    # calculate the average insolation the planet receives
    L = self.stellar_luminosity(distribution=distribution)
    a = self.semimajoraxis(distribution=distribution)
    S = L / 4 / np.pi / a**2
    return S.to(u.W / u.m**2)


def relative_insolation(self, distribution=False, **kw):
    """
    Relative Planet Insolation  (S/S_Earth)

    Calculate the insolation the planet receives from its star,
    given the luminosity of the star and the semimajor axis,
    expressed relative to Earth's insolation.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    return self.insolation(distribution=distribution) / earth_insolation


def log_relative_insolation(self, distribution=False, **kw):
    """
    log(Relative Planet Insolation)

    Calculate log10 of the insolation the planet receives from its star,
    given the luminosity of the star and the semimajor axis,
    expressed relative to Earth's insolation.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    return np.log10(self.relative_insolation(distribution=distribution))


def relative_cumulative_xuv_insolation(self, distribution=False, **kw):
    """
    Relative Cumulative XUV Insolation

    Calculate a *very very very* approximate estimate for the cumulative XUV flux (J/m**2)
    felt by a planet over its lifetime. It comes from Zahnle + Catling (2017) Equation 27,
    where they say they did integrals over the Lammer et al. (2009) XUV flux relations.
    This effectively assumes that the early times dominate, so the time integral doesn't
    depend (?!?) on the age of the system. It's very rough!

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    xuv_proxy = (
        self.stellar_luminosity(distribution=distribution).to_value(u.Lsun) ** -0.6
    )
    return self.relative_insolation(distribution=distribution) * xuv_proxy


def teq(self, distribution=False, albedo=0, f=1 / 4, **kw):
    """
    Planet Equilibrium Temperature (K)

    Calculate the equilbrium temperature of the planet. With default keywords,
    this assumes zero-albedo and uniform heat redistribution.

    Parameters
    ----------
    albedo : float
        The planet's Bond albedo, meaning the fraction of incoming
        stellar light reflected away, integrated over wavelength.
        The remaining light is absorbed and needs to be emitted
        as thermal (often infrared) radiation, thus contributing
        to the equilibrium temperature.
    f : float
        The ratio of the effective area collecting light from the star
        (which is the cross sectional area of the planet's shadow pi*R**2)
        to the effective area emitting thermal radiation away to space
        (which depends on how efficiently the planet distributes heat).
        Higher values of f correspond to a smaller fraction of the planet's
        surface doing the emitting, and thus a higher temperature.
        Two limiting values are:
            f = 1/4 = planet uniformly distributes heat, thus radiating
                      from a spherical surface with 4*pi*R**2 area
            f = 2/3 = looking at the dayside of planet that instantly
                      reradiates absorbed heat, with stronger weighting
                      toward the substellar point meaning the effective
                      emission area is only 3/2*pi*R**2 (which is smaller
                      than the spherical area of the dayside hemisphere,
                      which would give f=1/2)
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    S = self.insolation(distribution=distribution)
    sigma = con.sigma_sb
    teq = ((S * f * (1 - albedo) / sigma) ** (1 / 4)).to(u.K)
    return teq


def planet_luminosity(self, distribution=False, **kw):
    """
    Planet Luminosity (K)

    Calculate total thermal luminosity of the planet,
    due to reradiation of absorbed starlight.

    Parameters
    ----------
    albedo : float
        The planet's Bond albedo, meaning the fraction of incoming
        stellar light reflected away, integrated over wavelength.
        The remaining light is absorbed and needs to be emitted
        as thermal (often infrared) radiation, thus contributing
        to the equilibrium temperature.
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    S = self.insolation(distribution=distribution)
    R = self.radius(distribution=distribution)
    L_in = (S * np.pi * R**2).to(u.W)
    L_out = L_in
    return L_out


def transit_depth_from_radii(self, distribution=False, **kw):
    """
    Transit Depth ((Rp/Rs)**2, unitless)

    Calculate the transit depth from the planet and star radius.
    This calculates simply (Rp/Rs)**2; it neglects the effects of
    limb-darkening and/or grazing transits.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """

    Rp = self.radius(distribution=distribution)
    Rs = self.stellar_radius(distribution=distribution)
    depth = (Rp / Rs).decompose() ** 2
    return depth


def transit_depth(self, distribution=False, **kw):
    """
    Transit Depth ((Rp/Rs)**2, unitless)

    Retrieve transit depth first from the standardized table,
    then calculated from planet radius + stellar radius.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    return self._choose_calculation(
        methods=[
            "transit_depth_from_table",
            "transit_depth_from_radii",
        ],
        distribution=distribution,
        **kw,
    )


def scaled_radius_from_radii(self, distribution=False, **kw):
    """
    Scaled Planet Radius ((Rp/Rs), unitless)

    Calculate the radius ratio from the planet and star radius.
    This simply calculates (Rp/Rs); it might be used if no
    radius ratio is provided in the table.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """

    Rp = self.radius(distribution=distribution)
    Rs = self.stellar_radius(distribution=distribution)
    ratio = (Rp / Rs).decompose()
    return ratio


def scaled_radius(self, distribution=False, **kw):
    """
    Scaled Planet Radius ((Rp/Rs), unitless)

    Retrieve transit depth first from the standardized table,
    then calculated from planet radius + stellar radius.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    return self._choose_calculation(
        methods=[
            "transit_scaled_radius_from_table",
            "scaled_radius_from_radii",
        ],
        distribution=distribution,
        **kw,
    )


def transit_duration_from_orbit(self, distribution=False, **kw):
    """
    Transit Duration (days)

    Calculate the total transit duration (= the time between 1st and
    4th contact, from when first touches the stellar disk to when it
    last leaves). This calculation might be used if there is no
    transit duration available in the standardized table.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        P = self.period(distribution=distribution)
        scaled_semimajoraxis = self.scaled_semimajoraxis(distribution=distribution)
        b = self.transit_impact_parameter(distribution=distribution)
        k = self.scaled_radius(distribution=distribution)
        T0 = P / np.pi / scaled_semimajoraxis
        T = T0 * np.sqrt((1 + k**2) - b**2)

        e = self.eccentricity(distribution=distribution)
        omega = self.argument_of_periastron(distribution=distribution)
        factor = np.sqrt(1 - e**2) / (1 + e * np.sin(omega))

        duration = (T * factor).to(u.day)

    return duration


def transit_duration(self, distribution=False, **kw):
    """
    Transit Duration (days)

    Retrieve transit duration first from the standardized table,
    then calculated from the orbital parameters

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    return self._choose_calculation(
        methods=[
            "transit_duration_from_table",
            "transit_duration_from_orbit",
        ],
        distribution=distribution,
        **kw,
    )


def mass_estimated_from_radius(self, distribution=False, **kw):
    """
    Estimated Planet Mass (Earth masses)

    Retrieve an estimate of the planet mass from the
    planet's radius, using a very approximate
    empirical relation.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.

    """
    if distribution:
        raise RuntimeError(
            "mass_estimated_from_radius does not propagate uncertainties yet; sorry!"
        )
    radius = self.get("radius", distribution=False)
    return use_chen_and_kipping_to_estimate_mass_from_radius(R=radius)


def radius_estimated_from_mass(self, distribution=False, **kw):
    """
    Estimated Planet Radius (Earth radii)

    Retrieve an estimate of the planet radius from the
    planet's mass, using a very approximate
    empirical relation.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.

    """
    if distribution:
        raise RuntimeError(
            "radius_estimated_from_mass does not propagate uncertainties yet; sorry!"
        )

    mass = self.get("mass", distribution=False)
    return use_chen_and_kipping_to_estimate_radius_from_mass(M=mass)


def kludge_mass(self, distribution=False, **kw):
    """
    Planet Mass or msini (Earth masses)

    Retrieve an estimate of the planet mass,
    starting first from an actual published table mass,
    and then from msini assuming sini=1,
    and then from an emprical radius-mass relation.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.

    """
    return self._choose_calculation(
        methods=["mass_from_table", "msini_from_orbit", "mass_estimated_from_radius"],
        distribution=distribution,
        **kw,
    )


def kludge_radius(self, distribution=False, **kw):
    """
    Planet Radius or Estimated Planet Radius (Earth radii)

    Retrieve an estimate of the planet radius,
    starting first from an actual published table radius,
    and then from a mass-radius relation.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.

    """
    return self._choose_calculation(
        methods=[
            "radius_from_table",
            "radius_estimated_from_mass",
        ],
        distribution=distribution,
        **kw,
    )


def kludge_stellar_age(self, distribution=False, **kw):
    """
    System Age (Gyr)

    Retrieve an estimate of the system age,
    starting first from an actual published table age,
    and then boldy/foolishly assuming it's 5 Gyr.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.

    """

    # pull out the actual values from the table
    age = self.get("stellar_age", distribution=distribution)

    # try to replace bad ones with NVK3L
    bad = np.isfinite(age) == False
    age[bad] = 5 * u.Gyr

    return age


def surface_gravity(self, kludge=False, distribution=False, **kw):
    """
    Planet Surface Gravity (m/s**2)

    Calculate the planet's surface gravity
    from its mass and radius.

    TO-DO:
    Explore if we should choose a different calculation more
    closely tied to observables to minimize uncertainties.

    Parameters
    ----------
    kludge : bool
        Should we include kludged estimates for mass (from msini and/or
        empirical mass-radius) and/or radius (from empircal mass-radius)
        when doing this calculation?
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """

    # choose whether or not include estimated masses/radii
    if kludge:
        M = self.get("kludge_mass", distribution=distribution)
        R = self.get("kludge_radius", distribution=distribution)
    else:
        M = self.get("mass", distribution=distribution)
        R = self.get("radius", distribution=distribution)

    G = con.G
    g = (G * M / R**2).to("m/s**2")
    return g


def density(self, kludge=False, distribution=False, **kw):
    """
    Planet Density (m/s**2)

    Calculate the planet's bulk density
    from its mass and radius.

    TO-DO:
    Explore if we should choose a different calculation more
    closely tied to observables to minimize uncertainties.

    Parameters
    ----------
    kludge : bool
        Should we include kludged estimates for mass (from msini and/or
        empirical mass-radius) and/or radius (from empircal mass-radius)
        when doing this calculation?
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """

    # choose whether or not include estimated masses/radii
    if kludge:
        M = self.get("kludge_mass", distribution=distribution)
        R = self.get("kludge_radius", distribution=distribution)
    else:
        M = self.get("mass", distribution=distribution)
        R = self.get("radius", distribution=distribution)

    V = 4 / 3 * np.pi * R**3
    density = (M / V).to("g/cm**3")
    return density


def escape_velocity(self, kludge=False, distribution=False, **kw):
    """
    Planet Escape Velocity (km/s)

    Calculate the planet's escape velocity
    from its mass and radius.

    TO-DO:
    Explore if we should choose a different calculation more
    closely tied to observables to minimize uncertainties.

    Parameters
    ----------
    kludge : bool
        Should we include kludged estimates for mass (from msini and/or
        empirical mass-radius) and/or radius (from empircal mass-radius)
        when doing this calculation?
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """

    if kludge:
        M = self.get("kludge_mass", distribution=distribution)
        R = self.get("kludge_radius", distribution=distribution)
    else:
        M = self.get("mass", distribution=distribution)
        R = self.get("radius", distribution=distribution)

    G = con.G
    escape_velocity = np.sqrt(2 * G * M / R).to("km/s")
    return escape_velocity


def orbital_velocity(self, distribution=False, **kw):
    """
    Planet Orbital Velocity (km/s)

    Calculate the average tangential orbital speed of the
    the planet in its orbit around its star.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    a = self.semimajoraxis(distribution=distribution)
    P = self.period(distribution=distribution)
    return (2 * np.pi * a / P).to("km/s")


def impact_velocity(self, distribution=False, **kw):
    """
    Very Approximately Estimated Impact Velocity (km/s)

    Calcuate a back-of-the-envelope estimate of the typical
    impact velocity which which something might hit the planet,
    as a combination of orbital and escape velocity.

    Parameters
    ----------
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    v_orbital = self.orbital_velocity(distribution=distribution)
    v_escape = self.escape_velocity(distribution=distribution)
    return np.sqrt(v_orbital**2 + v_escape**2)


def escape_parameter(
    self,
    temperature="teq",
    mu=1,
    albedo=0,
    f=1 / 4,
    kludge=False,
    distribution=False,
    **kw
):
    """
    Jeans Atmospheric Escape Parameter (unitless)

    Calculate the Jeans escape parameter, as the ratio
    of a particle's gravitational binding energy to its
    thermal energy. This is a super-approximate tracer
    for susceptability to atmospheric loss, depending on
    lots of quantities that we often in truth don't know.

    TO-DO:
    Think about whether there is an interesting intermediate
    to assume for exospheric temperature, or whether we should
    just keep this as a simple relative scaling.

    TO-DO:
    Should we include a correction factor for different
    gravitational acceleration at high altitudes?

    Parameters
    ----------
    temperature : str, Quantity
        The temperature to use for the planet's exosphere,
        in units of Kelvin. The default string `temperature='teq'`
        will calculate an equilibrium temperature (assuming some
        albedo and heat redistribution parameter); otherwise
        something like `temperature=1000*u.K` will assume a
        constant exospheric temperature. Neither approximation
        is probably super great.
    mu : float
        Mean molecular weight of escaping particle,
        in atomic mass units. The default of mu=1
        corresponds to atomic hydrogen.
    albedo : float
        Albedo for calculating equilibrium temperature
        (see docstring for `.teq()` for details)
    f : float
        Heat redistribution for calculating equilibrium temperature
        (see docstring for `.teq()` for details)
    kludge : bool
        Should we include kludged estimates for mass (from msini and/or
        empirical mass-radius) and/or radius (from empircal mass-radius)
        when doing this calculation?
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    if temperature == "teq":
        T = self.teq(albedo=albedo, f=f, distributio=distribution)
    else:
        T = temperature

    if kludge:
        M = self.get("kludge_mass", distribution=distribution)
        R = self.get("kludge_radius", distribution=distribution)
    else:
        M = self.get("mass", distribution=distribution)
        R = self.get("radius", distribution=distribution)

    k = con.k_B
    m_p = con.m_p
    G = con.G
    e_thermal = k * T
    e_grav = G * M * m_p / R
    return (e_grav / e_thermal).decompose()


def scale_height(self, mu=2.3, albedo=0, f=1 / 4, kludge=False, distribution=False):
    """
    Atmospheric Scale Height (km)

    Calculate the scale height of the planet's atmosphere,
    assuming a mean molecular weight and the

    TO-DO:
    Consider alternate ways of estimating mean-molecular weight
    as a function of planet properties, so that (for example),
    we could use this to calculate scale heights for primary
    and secondary atmospheres together. There are probably
    so many unknowns that would need to go into this that
    we should make people makes those choices themselves.

    Parameters
    ----------
    mu : float
        Mean molecular weight of the atmosphere,
        in atomic mass units. The default of mu=2.3
        corresponds approximately to chemical equilibrium
        of a solar composition atmosphere at fairly
    albedo : float
        Albedo for calculating equilibrium temperature
        (see docstring for `.teq()` for details)
    f : float
        Heat redistribution for calculating equilibrium temperature
        (see docstring for `.teq()` for details)
    kludge : bool
        Should we include kludged estimates for mass (from msini and/or
        empirical mass-radius) and/or radius (from empircal mass-radius)
        when doing this calculation?
    distribution : bool
        If False, return a simple array of values.
        If True, return an astropy.uncertainty.Distribution,
        which can be used for error propagation.
    """
    k = con.k_B
    T = self.teq(albedo=albedo, f=f, distribution=distribution)
    m_p = con.m_p
    g = self.surface_gravity(kludge=kludge, distribution=distribution)
    H = (k * T / mu / m_p / g).to("km")
    return H



from .population_core import *
from .predefined import PredefinedPopulation
from .exoplanets import *
from .solarsystem import *

from .summary import *
from .column_descriptions import *
from .collections import *



import numpy as np
import astropy.units as u


def update_planet(pop):
    # Grimm et al. masses
    pop.update_planet(
        "TRAPPIST-1b",
        mass=1.017,
        mass_uncertainty_upper=0.154,
        mass_uncertainty_lower=-0.143,
    )

    pop.update_planet(
        "TRAPPIST-1c",
        mass=1.156,
        mass_uncertainty_upper=0.142,
        mass_uncertainty_lower=-0.131,
    )

    pop.update_planet(
        "TRAPPIST-1d",
        mass=0.297,
        mass_uncertainty_upper=0.039,
        mass_uncertainty_lower=-0.035,
    )

    pop.update_planet(
        "TRAPPIST-1e",
        mass=0.772,
        mass_uncertainty_upper=0.079,
        mass_uncertainty_lower=-0.075,
    )

    pop.update_planet(
        "TRAPPIST-1f",
        mass=0.934,
        mass_uncertainty_upper=0.080,
        mass_uncertainty_lower=-0.078,
    )
    pop.update_planet(
        "TRAPPIST-1g",
        mass=1.148,
        mass_uncertainty_upper=0.098,
        mass_uncertainty_lower=-0.095,
    )
    pop.update_planet(
        "TRAPPIST-1h",
        mass=0.331,
        mass_uncertainty_upper=0.056,
        mass_uncertainty_lower=-0.049,
    )

    pop.update_planet(
        "GJ 1132b",
        a_over_r=16.54,
        mass=1.62,
        mass_uncertainty_upper=0.55,
        mass_uncertainty_lower=-0.55,
        radius=1.130,
        radius_uncertainty_upper=0.056,
        radius_uncertainty_lower=-0.056,
        stellar_teff=3270.0,
        b=0.38,
        stellar_radius=0.2105,
        stellar_mass=0.181,
        transit_midpoint=2457184.55786,
    )

    pop.update_planet("30 Ari Bb", radius=np.nan)

    pop.update_planet("HD 17156b", transit_duration=0.1338, a_over_r=23.2, b=0.4)

    pop.update_planet("HD 17156b", transit_duration=0.1338, a_over_r=23.2, b=0.477)

    pop.update_planet("HD 80606b", transit_duration=0.4958, a_over_r=94, b=0.75)

    pop.update_planet("GJ 1132b", transit_duration=47.0 / 60.0 / 24.0)

    pop.update_planet("GJ 1214b", transit_duration=0.03620)

    pop.update_planet("GJ 436b", transit_duration=0.04227)

    pop.update_planet(
        "HD 219134b",
        a_over_r=1 / (0.03876 * u.au / 0.778 / u.Rsun).decompose().value,
        mass=4.74,
        mass_uncertainty_upper=0.19,
        mass_uncertainty_lower=-0.19,
    )

    pop.update_planet(
        "HD 219134c",
        a_over_r=1 / (0.06530 * u.au / 0.778 / u.Rsun).decompose().value,
        mass=4.36,
        mass_uncertainty_upper=0.22,
        mass_uncertainty_lower=-0.22,
    )

    pop.update_planet("LHS 1140b", transit_duration=2.0 / 24.0, a_over_r=101.0, b=0.16)

    pop.update_planet("GJ 3470b", transit_duration=0.07992)

    pop.update_planet("Kepler-51b", transit_duration=0.24154)

    pop.update_planet("Kepler-51c", transit_duration=0.1171)

    pop.update_planet("Kepler-51d", transit_duration=0.351992)

    pop.update_planet("Kepler-42b", transit_duration=0.022329)

    pop.update_planet("Kepler-42c", transit_duration=0.02)

    pop.update_planet("Kepler-42d", transit_duration=0.016554)

    pop.update_planet("WASP-43b", transit_duration=0.0483, a_over_r=4.8, b=0.66)

    pop.update_planet("K2-18b", transit_duration=0.1117)

    """I corrected the famous ones, which have weird choices in the archive."""
    pop.update_planet(
        "HD 209458b",
        mass=235.0,
        mass_uncertainty_upper=19.0,
        mass_uncertainty_lower=-19.0,
        stellar_teff=6065.0,
        a_over_r=8.9,
        b=0.5,
    )

    pop.update_planet(
        "HD 189733b",
        a_over_r=8.9,
        radius=13.630,
        radius_uncertainty_upper=0.269,
        radius_uncertainty_lower=-0.269,
        mass=369.303,
        mass_uncertainty_upper=18.433,
        mass_uncertainty_lower=-18.433,
        transit_midpoint=2454279.436714,
    )

    """ I went through all the planets shown in the Dressing et al. (2015)
        mass-radius diagram to double check the values that are present in the
        NASA Exoplanet Archive."""

    # parameters taken from the updates in Demory et al. (2015, *submitted*)
    """pop.update_planet('55 Cnce',
                radius=1.92,
                radius_uncertainty_upper=0.08,
                radius_uncertainty_lower=-0.08,
                a_over_r=3.53,
                mass=8.08,
                mass_uncertainty_upper=0.31,
                mass_uncertainty_lower=-0.31,
                b=0.36)"""

    # from Gillon et al. (2012) (and some from Demory et al.)
    pop.update_planet(
        "55 Cnce",
        radius=2.17,
        radius_uncertainty_upper=0.1,
        radius_uncertainty_lower=-0.1,
        a_over_r=3.53,
        mass=8.08,
        mass_uncertainty_upper=0.31,
        mass_uncertainty_lower=-0.31,
        b=0.459,
    )

    # from von Braun et al. (2012)
    pop.update_planet("GJ 436b", stellar_teff=3416.0)

    # from the updates in Anglada-Escude et al. (2013)
    pop.update_planet(
        "GJ 1214b",
        stellar_teff=3252.0,
        stellar_radius=0.211,
        radius=2.72,
        radius_uncertainty_upper=0.24,
        radius_uncertainty_lower=-0.24,
        a_over_r=14.62,
        rv_semiamplitude=10.9,
        mass=6.19,
        mass_uncertainty_upper=0.91,
        mass_uncertainty_lower=-0.91,
        b=0.2,
        distance=14.55,
        distance_uncertainty_upper=0.13,
        distance_uncertainty_lower=-0.13,
    )

    # from Knutson et al. (2014)
    pop.update_planet("HD 97658b", a_over_r=26.24)

    # from Haywood et al. (2014), allowing eccentricity as they do in abstract
    pop.update_planet(
        "CoRoT-7b",
        rv_semiamplitude=3.42,
        mass=4.73,
        mass_uncertainty_upper=0.95,
        mass_uncertainty_lower=-0.95,
    )

    # from Carter et al.
    pop.update_planet("Kepler-36b", a_over_r=15.24)

    # the archive parameters look reasonable for
    # HIP 116454b
    # Kepler-10b
    # Kepler-10c
    # Kepler-93b

    """ Then, I corrected things that popped up as weird non-physical outliers
        in various slices of parameter space, trying to track down real values
        whenever possible."""

    # Kepler-11's radius was listed as 0.07!
    pop.update_planet("Kepler-11b", stellar_radius=1.053)
    pop.update_planet("Kepler-11c", stellar_radius=1.053)
    pop.update_planet("Kepler-11d", stellar_radius=1.053)
    pop.update_planet("Kepler-11e", stellar_radius=1.053)
    pop.update_planet("Kepler-11f", stellar_radius=1.053)
    pop.update_planet("Kepler-11g", stellar_radius=1.053)

    # googling the ones that have messed up stellar_teffs
    pop.update_planet("Qatar-1b", stellar_teff=4861.0)
    pop.update_planet("Kepler-450b", stellar_teff=6215.0)
    pop.update_planet("Kepler-450c", stellar_teff=6215.0)
    pop.update_planet("Kepler-450d", stellar_teff=6215.0)

    pop.update_planet("Kepler-449b", stellar_teff=5588.0)
    pop.update_planet("Kepler-449c", stellar_teff=5588.0)

    pop.update_planet("Kepler-414b", stellar_teff=5523.0)
    pop.update_planet("Kepler-414c", stellar_teff=5523.0)

    pop.update_planet("Kepler-415b", stellar_teff=4523.0)
    pop.update_planet("Kepler-415c", stellar_teff=4523.0)

    pop.update_planet("Kepler-416b", stellar_teff=5670.0)
    pop.update_planet("Kepler-416c", stellar_teff=5670.0)

    pop.update_planet("Kepler-417b", stellar_teff=5376.0)
    pop.update_planet("Kepler-417c", stellar_teff=5376.0)

    # from discovery paper
    pop.update_planet("WASP-100b", stellar_radius=2.0)

    # supposedly HIPPARCOS, via exoplanets.org
    pop.update_planet(
        "HD80606b",
        distance=58,
        distance_uncertainty_upper=30,
        distance_uncertainty_lower=-14.8,
    )

    # from discovery paper
    pop.update_planet("Kepler-128b", mass=np.nan)
    pop.update_planet("Kepler-128c", mass=np.nan)



from .setup_tests import *

from exoatlas import *
from exoatlas.visualizations import *


def test_reflection(telescope_name="JWST", wavelength=1 * u.micron):
    with mock.patch("builtins.input", return_value=""):
        t = TransitingExoplanets()

    w = 1 * u.micron
    fi, ax = plt.subplots(1, 2, figsize=(8, 4), constrained_layout=True)
    for i, per_transit in enumerate([True, False]):
        BubbleMap(
            xaxis=StellarBrightnessTelescope(
                wavelength=wavelength, telescope_name=telescope_name
            ),
            yaxis=Reflection(wavelength=w),
            size=ReflectionSNR(
                wavelength=wavelength,
                telescope_name=telescope_name,
                per_transit=per_transit,
            ),
            size_normalization=10,
        ).build(t, ax=ax[i])
        d = {True: "Transit", False: "Hour"}
        plt.title(f"S/N in 1 {d[per_transit]}")


def test_emission(telescope_name="JWST", wavelength=5 * u.micron):
    with mock.patch("builtins.input", return_value=""):
        t = TransitingExoplanets()

    w = 5 * u.micron
    fi, ax = plt.subplots(1, 2, figsize=(8, 4), constrained_layout=True)
    for i, per_transit in enumerate([True, False]):
        BubbleMap(
            xaxis=StellarBrightnessTelescope(
                wavelength=wavelength, telescope_name=telescope_name
            ),
            yaxis=Emission(wavelength=wavelength),
            size=EmissionSNR(wavelength=wavelength, telescope_name=telescope_name),
            size_normalization=10,
        ).build(t, ax=ax[i])
        d = {True: "Transit", False: "Hour"}
        plt.title(f"S/N in 1 {d[per_transit]}")


def test_transmission(telescope_name="JWST", wavelength=5 * u.micron):
    with mock.patch("builtins.input", return_value=""):
        t = TransitingExoplanets()

    w = 5 * u.micron
    fi, ax = plt.subplots(1, 2, figsize=(8, 4), constrained_layout=True)
    for i, per_transit in enumerate([True, False]):
        BubbleMap(
            xaxis=StellarBrightnessTelescope(
                wavelength=wavelength, telescope_name=telescope_name
            ),
            yaxis=Transmission(),
            size=TransmissionSNR(wavelength=wavelength, telescope_name=telescope_name),
            size_normalization=10,
        ).build(t, ax=ax[i])
        d = {True: "Transit", False: "Hour"}
        plt.title(f"S/N in 1 {d[per_transit]}")


if __name__ == "__main__":
    outputs = {k.split("_")[-1]: v() for k, v in locals().items() if "test_" in k}



from .setup_tests import *

from exoatlas.imports import *


def test_directory():
    print(directories["data"])


def test_name2color():
    one = name2color("black")
    another = name2color("#000000")
    assert one == another


if __name__ == "__main__":  # pragma: no cover
    outputs = {k.split("_")[-1]: v() for k, v in locals().items() if "test_" in k}



from ...imports import *
from ...populations import Population
from matplotlib.colors import Normalize, LogNorm


class Plottable:
    source = None
    label = None
    scale = "log"
    lim = [None, None]
    unit = None
    """
    General class definition for a plottable axis,
    which can appear as either an xaxis or yaxis.

    The standard items to include in the definition
    of a plottable object are:


    """

    def __init__(self, source=None, label=None, scale=None, lim=None, unit=None, **kw):
        """
        Initialize a Plottable quantity.

        This `Plottable` serves as a helper for plotting
        `exoatlas` quantities. It provides functions to
        return values and uncertainties for a `Population`,
        but it also stores important information about
        labels, scales, and limits that are needed for
        good automatic visualizations (but are not needed
        for simply doing calculations).

        Often a `Plottable` will be connected to some kind
        of a visualization `Map`, which will handle rendering
        data on plots and cycling through populations.

        Parameters
        ----------
        source : str
            Generally a string, indicating a
            retrievable population attribute,
            which will be used as the value
            to plot. The values-getting can
            also be over-written by defining
            a .value() method, if the desired
            values are more complicated than
            a simple attribute/property (e.g.
            if it depends on an input, like
            wavelength).
        label : str
            Human-friendly string describing
            this axis, which will appear as
            its label in the plot.
        scale : str
            String ('linear', 'log', ?) to
            indicate what scale should be
            used for this axis.
        lim : tuple
            Tuple to indicate (lower, upper)
            limits for the plot. These might
            also be used to set the vmin and
            vmax of a color map, if this
            plottable is being used to color
            points in a BubbleMap. `lim`
            should have no units attached.
            For compariing different populations,
            it's probably a very good idea to
            explicitly set the limits and not
            leave them as None; otherwise
            individual populations might normalize
            themselves separately.
        unit : astropy.units.Unit
            The astropy unit to be used for plotting.
            It must be something to which the
            quantity can be converted.
        kw : dict
            All other keywords will be passed to .get()
            when retrieving values or uncertainties.
            For example, for source='teq', we might
            consider passing 'albedo' or 'f' keywords,
            to therefore get the values and uncertainties
            associated with a particular parameters.
        """

        # set properties to be defaults, updated by keywords
        self.source = source or self.source
        self.label = label or self.label
        self.scale = scale or self.scale
        self.lim = lim or self.lim
        self.unit = unit or self.unit

        # set keywords to be defaults for source, updated by **kw
        self.kw = self._get_default_keyword_arguments() | kw

        # after .source and .kw have been defined, update label, if necessary
        self._update_label()

    def _update_label(self):
        """
        Update the label.

        Sometimes, we might not know everything
        we need to write an appropriate label
        until after the `Plottable` has been
        created. For example, if a quantity is
        wavelength-dependent, we need to make
        an instance of the `Plottable` with a
        real wavelength, before we know how to
        set the wavelength label. If so,
        this method should be overwritten
        when defining a new `Plottable` class.

        When called, it should change
        `self.label`, and it can count on
        `self.source` and `self.kw` already
        being defined.
        """
        pass

    def _get_default_keyword_arguments(self):
        """
        Figure out default science keyword arguments.

        For the `source` method that returns the quantities
        to be plotted, figure out what the default keyword
        arguments are, so if this `Plottable` gets created
        without custom keyword values, we can still
        make use of the keyword inputs.
        """
        try:
            # get method for calculating quantity
            f = getattr(Population, self.source)
        except AttributeError:
            # Table columns that don't have explicit
            # method definitions probably won't show
            # up in the Population class until after
            # an instance is created. That's OK, because
            # those won't have custom keywords we need
            # care about.
            return {}

        # figure out default keywords
        kw = {
            k: v.default
            for k, v in inspect.signature(f).parameters.items()
            if v.default is not inspect.Parameter.empty
        }
        try:
            kw.pop("distribution")
        except KeyError:
            pass
        return kw

    def __repr__(self):
        """
        How should this plottable axis be represented as a string?
        """
        return f"📏 {self.source or self.label or '?'}".replace("\n", " ")

    def _convert_unit(self, x):
        """
        Convert quantity into desired unit,
        unless units aren't specified.

        Parameter
        ---------
        x : quantity
            The quantity to convert to `self.unit`
        """
        if self.unit is None:
            return x
        else:
            try:
                return x.to(self.unit)
            except AttributeError:
                return x

    def value(self, pop):
        """
        Extract the values for a population.

        Generally, plotting maps might loop
        through multiple populations to include
        each on the plot.

        By default, this is done by using the
        string in `self.source` to pull results
        from a `Population` method. Any keywords
        stored in `self.kw` will also be passed
        to the call for that method.

        Write over this function in order to make
        more complicated function calls, if necessary.

        Parameters
        ----------
        pop : Population
            The `exoatlas` population for which
            values will be retreived.

        Returns
        -------
        value : Quantity
            The value an astropy Quantity array.
        """
        value = pop.get(self.source, **self.kw)
        return self._convert_unit(value)

    def uncertainty_lowerupper(self, pop):
        """
        Extract the upper and lower uncertainties for a population.

        Generally, plotting maps might loop
        through multiple populations to include
        each on the plot.

        By default, this is done by using the
        string in `self.source` to pull uncertainties
        from a `Population` method. Any keywords
        stored in `self.kw` will also be passed
        to the call for that method.


        You might want to overwrite this method for
        quantities where the uncertainty returned
        by the Population's normal uncertainty
        propagation is, for whatever reason, not
        exactly what you want.

        Parameters
        ----------
        pop : Population
            The `exoatlas` population for which
            values will be retreived.

        Returns
        -------
        lower : np.array
            The magnitude of the lower uncertainties (x_{-lower}^{+upper}),
            as an astropy Quantity array.
        upper : np.array
            The magnitude of the upper uncertainties (x_{-lower}^{+upper}),
            as an astropy Quantity array.
        """
        lower, upper = pop.get_uncertainty_lowerupper(self.source, **self.kw)
        return self._convert_unit(lower), self._convert_unit(upper)

    def uncertainty(self, pop):
        """
        Extract the uncertainties for a population.

        Generally, plotting maps might loop
        through multiple populations to include
        each on the plot.

        By default, this is done by using the
        string in `self.source` to pull uncertainties
        from a `Population` method. Any keywords
        stored in `self.kw` will also be passed
        to the call for that method.

        You might want to overwrite this method for
        quantities where the uncertainty returned
        by the Population's normal uncertainty
        propagation is, for whatever reason, not
        exactly what you want.

        Parameters
        ----------
        pop : Population
            The `exoatlas` population for which
            values will be retreived.

        Returns
        -------
        sigma : np.array
            The uncertainties an astropy Quantity array.
        """
        sigma = pop.get_uncertainty(self.source, **self.kw)
        return self._convert_unit(sigma)

    def normalized_value(self, pop):
        """
        Extract the values for a population,
        normalized to fall between 0 and 1.

        This normalizes the values for a population
        according to scale

        Write over this function in order to make
        more complicated function calls, if necessary.

        Parameters
        ----------
        pop : Population
            The `exoatlas` population for which
            values will be retreived.

        Returns
        -------
        normalized_value : Quantity
            The value, normalized
        """

        # get unitless array of values
        value = remove_unit(self._convert_unit(pop.get(self.source, **self.kw)))

        # set non-nan limits if none are provided
        vmin, vmax = self.lim
        if vmin == None:
            vmin = np.nanmin(value)
        if vmax == None:
            vmax = np.nanmax(value)

        # call different scaling functions
        reversed = vmin > vmax
        if reversed:
            kw = dict(vmin=vmax, vmax=vmin)
        else:
            kw = dict(vmin=vmin, vmax=vmax)
        if self.scale == "linear":
            normalize = Normalize(**kw)
        elif self.scale == "log":
            normalize = LogNorm(**kw)
        if reversed:
            return 1 - normalize(value)
        else:
            return normalize(value)


def clean_plottable(initial, **kw):
    """
    Make sure something is actually a Plottable object.

    We allow fairly flexible calls to plotting maps,
    where it's possible that a Plottable might be
    requested using a variety of different formats.
    This wrapper tries to make sure we get to what we
    need, a class definition, no matter what.

    (This used to accept a string to initialize a
    basic )

    Parameters
    ----------
    initial : Plottable, Plottable class, str, None
        One of a few ways to flexibly define a Plottable.
        Here's how the different options will be interpreted:

        Plottable =
            If already an instance, the Plottable object itself
            will be used for plotting. This would be common for
            something like `StellarBrightness(wavelength=5*u.micron)`,
            where keywords are necessary for what's being plotted.
        Plottable class =
            If a class definition, the Plottable will be
            created using the defaults for that class.
        None =
            The parent map probably has a Plottable
            connected to a particular axis; None just means
            not to overwrite that default.
    **kw : dict
        Extra keywords that should be passed in when
        initializing `Plottable` from a class or a string.
    Returns
    -------
    plottable_or_not : (various)
        Either an instance of a Plottable object,
        or something else entirely.
    """

    if isinstance(initial, Plottable):
        # pass through an actual Plottable object
        return initial
    elif type(initial) == type:
        if issubclass(initial, Plottable):
            # create an instance from a Plottable class, using defaults
            return initial(**kw)
    else:
        # pass others through
        return initial



from ..imports import *


def define_telescope_unit(
    telescope_name="telescope",
    area=1 * u.m**2,
    wavelength=5 * u.micron,
    R=20,
    dt=1 * u.hr,
    **kw,
):
    """
    Create a custom astropy unit to represent
    the collecting area of a generic telscope
    with a particular collecting area
    at a particular wavelength
    at a particular resolution
    observing for a particular time.

    Parameters
    ----------
    telescope_name : str
        The name of the telescope.

    area : astropy.units.quantity.Quantity
        The collecting area of the telescope.

    wavelength : astropy.unit.Quantity
        The wavelength at which it should be calculated.

    R : float
        The spectral resolution at which the
        telescope will bin wavelengths..
        (Ignored if telescope is None.)

    dt : astropy.units.quantity.Quantity
        The time over which the telescope exposes.
        (Ignored if telescope is None.)
    """

    dw = wavelength / R
    unit = u.def_unit(
        f"[{telescope_name} | {wavelength} | {dt} | R={R}]",
        dt * area * dw,
        doc=f"""
                      This custom unit represents the
                      photon-collecting power of the
                      {telescope_name} telescope, when integrating
                      for a time of {dt}, at a spectral
                      resolution of R={R} (evaluated
                      at {wavelength}).
                      """,
    )
    unit.telescope_name = telescope_name
    unit.wavelength = wavelength
    unit.dt = dt
    unit.R = R
    return unit


def define_JWST_unit(wavelength=5 * u.micron, efficiency=0.3, **kw):
    """
    Create a JWST telescope unit.

    Parameters
    ----------
    wavelength : astropy.unit.Quantity
        The wavelength at which it should be calculated.

    R : float
        The spectral resolution at which the
        telescope will bin wavelengths..
        (Ignored if telescope is None.)

    dt : astropy.units.quantity.Quantity
        The time over which the telescope exposes.
        (Ignored if telescope is None.)
    """
    return define_telescope_unit(
        telescope_name="JWST",
        wavelength=wavelength,
        area=efficiency * 25 * u.m**2,
        **kw,
    )


def define_HST_unit(wavelength=1.4 * u.micron, efficiency=0.3, **kw):
    """
    Create a HST telescope unit.

    Parameters
    ----------
    wavelength : astropy.unit.Quantity
        The wavelength at which it should be calculated.

    R : float
        The spectral resolution at which the
        telescope will bin wavelengths..
        (Ignored if telescope is None.)

    dt : astropy.units.quantity.Quantity
        The time over which the telescope exposes.
        (Ignored if telescope is None.)
    """
    return define_telescope_unit(
        telescope_name="HST",
        wavelength=wavelength,
        area=efficiency * 4.5 * u.m**2,
        **kw,
    )


def define_APO_unit(wavelength=0.5 * u.micron, efficiency=0.1, **kw):
    """
    Create an APO telescope unit.

    Parameters
    ----------
    wavelength : astropy.unit.Quantity
        The wavelength at which it should be calculated.

    R : float
        The spectral resolution at which the
        telescope will bin wavelengths..
        (Ignored if telescope is None.)

    dt : astropy.units.quantity.Quantity
        The time over which the telescope exposes.
        (Ignored if telescope is None.)
    """
    radius = 3.5 * u.m / 2
    area = np.pi * radius**2
    return define_telescope_unit(
        telescope_name="APO", wavelength=wavelength, area=area * efficiency, **kw
    )


def define_SBO_unit(wavelength=0.5 * u.micron, efficiency=0.1, **kw):
    """
    Create an APO telescope unit.

    Parameters
    ----------
    wavelength : astropy.unit.Quantity
        The wavelength at which it should be calculated.

    R : float
        The spectral resolution at which the
        telescope will bin wavelengths..
        (Ignored if telescope is None.)

    dt : astropy.units.quantity.Quantity
        The time over which the telescope exposes.
        (Ignored if telescope is None.)
    """
    radius = 0.5 * u.m / 2
    area = np.pi * radius**2
    return define_telescope_unit(
        telescope_name="SBO", wavelength=wavelength, area=area * efficiency, **kw
    )


def define_TESS_unit(wavelength=0.8 * u.micron, R=2, **kw):
    """
    Create a TESS telescope unit.

    Parameters
    ----------
    wavelength : astropy.unit.Quantity
        The wavelength at which it should be calculated.

    R : float
        The spectral resolution at which the
        telescope will bin wavelengths..
        (Ignored if telescope is None.)

    dt : astropy.units.quantity.Quantity
        The time over which the telescope exposes.
        (Ignored if telescope is None.)
    """
    return define_telescope_unit(
        telescope_name="TESS", wavelength=wavelength, area=0.0086 * u.m**2, R=R, **kw
    )


def define_Kepler_unit(wavelength=0.65 * u.micron, R=2, **kw):
    """
    Create a Kepler telescope unit.

    Parameters
    ----------
    wavelength : astropy.unit.Quantity
        The wavelength at which it should be calculated.

    R : float
        The spectral resolution at which the
        telescope will bin wavelengths..
        (Ignored if telescope is None.)

    dt : astropy.units.quantity.Quantity
        The time over which the telescope exposes.
        (Ignored if telescope is None.)
    """
    return define_telescope_unit(
        telescope_name="Kepler", wavelength=wavelength, area=0.708 * u.m**2, R=R, **kw
    )


# create a dictionary of unit factories
telescope_units = dict(
    Kepler=define_Kepler_unit,
    TESS=define_TESS_unit,
    HST=define_HST_unit,
    JWST=define_JWST_unit,
    APO=define_APO_unit,
    SBO=define_SBO_unit,
)


def define_telescope_unit_by_name(telescope_name, wavelength=None, **kw):
    """
    Wrapper to create a telescope unit
    by passing the telescope name.

    Parameters
    ----------
    telescope_name : str
        The name of the telescope.

    wavelength : astropy.unit.Quantity
        The wavelength at which it should be calculated.

    R : float
        The spectral resolution at which the
        telescope will bin wavelengths..
        (Ignored if telescope is None.)

    dt : astropy.units.quantity.Quantity
        The time over which the telescope exposes.
        (Ignored if telescope is None.)
    """
    if wavelength is None:
        return telescope_units[telescope_name](**kw)
    else:
        return telescope_units[telescope_name](wavelength=wavelength, **kw)


# define a useful big number of photons (=30ppm)
lotsofphotons_unit = u.def_unit("Gigaphotons", 1e9 * u.ph)



from .setup_tests import *

from exoatlas import *
from exoatlas.imports import *
from exoatlas.visualizations import *

import exoatlas as ex
import matplotlib.pyplot as plt
from exoatlas.visualizations.maps.preset_maps import preset_maps


def test_plottables():
    e = TransitingExoplanets()[:3]
    for k, v in preset_plottables.items():
        x = v()
        print(k)
        print(x.kw)
        print(x.label)
        print(x.value(e))
        print()


def test_maps():
    pops = {}
    pops["exo"] = TransitingExoplanets()
    pops["solarsystem"] = SolarSystem()
    for k, p in preset_maps.items():
        print(k, p)
        p().build(pops=pops)
        plt.title(f"Map={k}")


def test_map_types():
    pops = {}
    pops["solarsystem"] = SolarSystem()
    pops["exo"] = TransitingExoplanets()

    fr = Flux_x_Radius()
    fr.build(pops=pops)

    fr = BubbleMap(
        xaxis=Plottable(source="insolation"), yaxis=Plottable(source="radius")
    )
    fr.build(pops=pops)

    fr = ErrorMap(
        xaxis=Plottable(source="insolation"), yaxis=Plottable(source="radius")
    )
    fr.build(pops=pops)


def test_galleries():
    pops = {}
    pops["solarsystem"] = SolarSystem()
    FourPanelTransitGallery().build(pops)
    PlanetGallery().build(pops)


if __name__ == "__main__":  # pragma: no cover
    outputs = {k.split("_")[-1]: v() for k, v in locals().items() if "test_" in k}
